# -*- coding: utf-8 -*-

import pytest
from fempy.geometry.tools import *


class TestTools:
    def test_make3D(self):
        np.testing.assert_almost_equal(coors3D([1.]), [1, 0, 0])
        np.testing.assert_almost_equal(coors3D(1.), [1, 0, 0])
        np.testing.assert_almost_equal(coors3D([[1], [2.]]),
                                       [[1, 0, 0], [2, 0, 0]])
        with pytest.raises(ValueError):
            coors3D([1., 2, 3, 4])
            coors3D([[1., 2, 3, 4], [1., 2, 3, 4]])
            coors3D([[[1]]])

#    def test_contains_points(self):
#        polygon = [[0, 0, 0], [1, 1, 0], [0, 0, 1]]
#        points1 = [[0.2, 0.2, 0.2], [0.3, 0.3, 0.3]]
#        points2 = [[0.2, 0.2, -0.2], [0.3, 0.3, 0.3]]
#        # points3 = [[0.5, 0.5, 0.5]]
#        assert contains_points(polygon, points1).all()
#        assert not contains_points(polygon, points2).all()
#        assert contains_points(polygon, points2).any()

    def test_add_dimensions(self):
        a = np.random.rand(10, 2)
        b = np.zeros((10, 1))
        r = np.append(a, b, axis=1)
        assert np.allclose(add_dimensions(a), r)

        a = np.random.rand(10, 1)
        b = np.zeros((10, 2))
        r = np.append(a, b, axis=1)
        assert np.allclose(add_dimensions(a), r)

        a = np.random.rand(10, 3)
        assert np.allclose(add_dimensions(a), a)

    def test_contains(self):
        pytest.importorskip('shapely')
        polygon = [[0, 0, 0], [1, 1, 0], [0, 0, 1]]
        points1 = [[0.2, 0.2, 0.2], [0.3, 0.3, 0.3]]
        points2 = [[0.2, 0.2, -0.2], [0.3, 0.3, 0.3]]
        # points3 = [[0.5, 0.5, 0.5]]
        assert contains(polygon, points1)
        assert not contains(polygon, points2)
#        assert contains_points(polygon, points2)

    def test_disjoint(self):
        pytest.importorskip('shapely')
        polygon = [[0, 0, 0], [1, 1, 0], [0, 0, 1]]
        points1 = [[0.2, 0.2, 0.2], [0.3, 0.3, 0.3]]
        points2 = [[0.2, 0.2, -0.2], [0.3, 0.3, 0.3]]
        points3 = [[0.2, 0.2, -0.2], [-0.3, -0.3, 0.3]]
        assert not disjoint(polygon, points1)
        assert not disjoint(polygon, points2)
        assert disjoint(polygon, points3)
        # assert not contains_points(polygon, points3) # when on border

    def test_natural_spline(self):
        from fempy.geometry import Point, Line, Spline, PlaneSurface, Polygon
        coors = [[0, 0], [np.pi/2., 1], [np.pi, 0.31], [3*np.pi/2., 1],
                 [2*np.pi, 0], [3*np.pi/2., -1]]
        pts = [Point(c) for c in coors]
        s = Spline(pts, order=2)
        S = PlaneSurface([s, Line(pts[0], pts[-1])], 'srf', elsize=0.1)

        x, y = list(zip(*coors))
        sx = NaturalSpline(x)
        sy = NaturalSpline(y)
        ti = np.linspace(0, 1, 200)
        xi = sx.eval(ti)
        yi = sy.eval(ti)
        poly = list(zip(xi, yi))
        P = Polygon(poly)
#        (P+S).gmsh.preview()

    def test_natural_closed(self):
        from fempy.geometry import Point, Spline, PlaneSurface, Polygon
        coors = [[0, 0], [np.pi/2., 1], [np.pi, 0.31], [3*np.pi/2., 1],
                 [2*np.pi, 0], [3*np.pi/2., -1], [0, 0]]
        pts = [Point(c) for c in coors]
        pts[-1] = pts[0]
        s = Spline(pts, order=2)
        S = PlaneSurface([s], 'srf', elsize=0.1)

        x, y = list(zip(*coors))
        sx = NaturalSpline(x, closed=True)
        sy = NaturalSpline(y, closed=True)
        ti = np.linspace(0, 1, 200)
        xi = sx.eval(ti)
        yi = sy.eval(ti)
        poly = list(zip(xi, yi))
        P = Polygon(poly)
#        (P+S).gmsh.preview()

    def test_gmsh_spline(self):
        from fempy.geometry import Point, Line, Spline, PlaneSurface, Polygon
        coors = [[0, 0], [np.pi/2., 1], [np.pi, 0.31], [3*np.pi/2., 1],
                 [2*np.pi, 0], [3*np.pi/2., -1]]
        pts = [Point(c) for c in coors]
        s = Spline(pts, order=2)
        S = PlaneSurface([s, Line(pts[0], pts[-1])], 'srf', elsize=0.1)

        x, y = list(zip(*coors))
        sx = GmshSpline(x)
        sy = GmshSpline(y)
        ti = np.linspace(0, 1, 200)
        xi = sx.eval(ti)
        yi = sy.eval(ti)
        poly = list(zip(xi, yi))
        P = Polygon(poly)
#        (P+S).gmsh.preview()

    def test_gmsh_spline_closed(self):
        from fempy.geometry import Point, Spline, PlaneSurface, Polygon
        coors = [[0, 0], [np.pi/2., 1], [np.pi, 0.31], [3*np.pi/2., 1],
                 [2*np.pi, 0], [3*np.pi/2., -1], [0, 0]]
        pts = [Point(c) for c in coors]
        pts[-1] = pts[0]
        s = Spline(pts, order=2)
        S = PlaneSurface([s], 'srf', elsize=0.1)

        x, y = list(zip(*coors))
        sx = GmshSpline(x, closed=True)
        sy = GmshSpline(y, closed=True)
        ti = np.linspace(0, 1, 200)
        xi = sx.eval(ti)
        yi = sy.eval(ti)
        poly = list(zip(xi, yi))
        P = Polygon(poly)
#        (P+S).gmsh.preview()


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""
import pygmsh
from fempy.geometry import Gmsh

# Geometry and domain
pg = pygmsh.built_in.Geometry()
c1 = pg.add_circle([0., 0., 0.], 1., 0.25)
c2 = pg.add_circle([0., 0., 0.], 5., 1, holes=[c1])
t1, v1, l1 = pg.extrude(c1.plane_surface, [0, 0, 3.])
t2, v2, l2 = pg.extrude(c2.plane_surface, [0, 0, 3.])
g = Gmsh(pg, entities={3: {1: 'c1',
                           2: 'c2'},
                       2: {5:  'c1_bottom',
                           10: 'c2_bottom',
                           27: 'c1_top',
                           59: 'c2_top',
                           18: 'c1_side',
                           22: 'c1_side',
                           26: 'c1_side',
                           38: 'c2_side',
                           42: 'c2_side',
                           46: 'c2_side'
                           }
                       }
         )
g.preview()



# -*- coding: utf-8 -*-

import pytest
from fempy.geometry.shapes import *

class TestShapes:
    def test_rectangle(self):
        R = Rectangle()
        R.gmsh.msh
        R.gmsh.options.Mesh.CharacteristicLengthFactor = 1
#        R.gmsh.preview()

    def test_rectangle_periodic(self):
        R = Rectangle(periodic=True)
        R.points[0].elsize_factor=0.25
        R.gmsh.msh
#        R.gmsh.options.Mesh.CharacteristicLengthFactor = 0.5
#        R.gmsh.preview()

    def test_rectangles(self):
        Rectangles().gmsh.msh
        #Rectangles().gmsh.preview()

    def test_polygon(self):
        p = Polygon([[0, 0], [1, 0],[0, 1]])
        p.gmsh.msh
        #p.gmsh.preview()

    def test_circle(self):
        Circle().gmsh.msh
        #Circle().gmsh.preview()

    def test_circles(self):
        Circles().gmsh.msh
        #Circles().gmsh.preview()

    def test_ellipse(self):
        Ellipse().gmsh.msh
        #Ellipse().gmsh.preview()

    def test_ellipses(self):
        E = Ellipses([(1, 0.5), (0.75, 0.4), (0.5, 0.25)])
        E.gmsh.msh
        #E.gmsh.options.Mesh.CharacteristicLengthFromCurvature = 1
        #E.gmsh.preview()

    def test_random_hull(self):
        pytest.importorskip('shapely')
        RandomHull().gmsh.msh
        #RandomHull().gmsh.preview()

    def test_flow_rve(self):
        FlowRVE().gmsh.msh
        #FlowRVE().gmsh.preview()

    def test_box(self):
        Box().gmsh.msh
        # Box().gmsh.preview()
        
    def test_cylinder(self):
        pytest.importorskip("pygmsh")
        g = Cylinder()
        g.gmsh.msh
        g.gmsh.domain()
#        g.preview()

    def test_multipy(self):
        R1 = Rectangle()
        R2 = R1.copy()
        R2.move([1, 0])
        R = R1 + R2
#        R.gmsh.preview()

    def test_save_load(self):
        import tempfile
        tempdir = tempfile.gettempdir()
        R = Rectangle()
        R.save(tempdir + '/aaabbbccc')
        R2 = g.Geometry.load(tempdir + '/aaabbbccc')
        R2.coors
        R2.gmsh.msh


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:54:09 2016

@author: mwojc


MISSING FUNCTIONALITIES AND TESTS

* only 'lop' should be sufficient to define model
* jacobian is a dense matrix?
* bcs in model changed?
* initialize solvers without model, handle model change properly
* possible lop types should be: method(x), property, LinearOperator,
  only property returning result is supported now
* possible jac types should be: method(x), property, LinearOperator, matrix,
  currently jac can be a sparse matrix or property returning it
* possible jaclm types should be as for jac
* it should be possible to define loplm instead of jaclm (be aware that
  loplm must implement B.T*x AND B*lm)
* test for custom 'ilop'
* test for custom 'ijac'


To be verified: maybe the 'method' in linear solvers should be used
inside M (in matvec defining M) on the reduced system instead of the
global use in 'solve' for the whole problem. That should work better
because the reduced problem is first smaller, secondly usually symmetric.
The problem is possibly the longer run whem lm are present...

"""
from __future__ import print_function
from fempy.compat import reload
import pytest
import inspect
import numpy as np
import scipy.sparse as sparse
import fempy.solvers.methods as methods
import fempy.solvers.linear as linear
import fempy.solvers.precon as precon
# import fempy.solvers.nonlin as nonlin


def get_solvers():
    reload(methods)
    reload(precon)
    reload(linear)
    mc = [ls for ls in methods.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, methods.linear)]

    pc = [ls for ls in precon.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, precon.preconditioner)]

    lsc = [ls for ls in linear.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, linear.LinearSolver)]
    return mc, pc, lsc

#    nsc = [ls for ls in nonlin.__dict__.values()
#           if inspect.isclass(ls) and issubclass(ls, nonlin.NonLinearSolver)]

# Preconditioners causing problems
pblacklist = {  # linear.CgsSolver: (None, precon.diag),
                # linear.CgSolver: (None, precon.diag),
                # linear.BicgstabSolver: (None, precon.diag),
                # linear.AmgBicgstabSolver: (None, precon.diag),
                # linear.AmgMinimalResidualSolver: (None, precon.diag),
                # linear.AmgGmresSolver: (None, precon.diag),
                # linear.AmgFgmresSolver: (None, precon.diag),
                # linear.GmresSolver: (precon.diag,),
              }

# Solvers causing problems
blacklist = [  # linear.MinresSolver,
               # linear.AmgSteepestDescentSolver,
               # linear.AmgCgSolver,
             ]


# Dummy model
class M:
    info = "MODEL 1 - no bcs, no lm"

    def __init__(self, kind=''):
        n = 10
        rhs = np.random.rand(n)
        jac = sparse.csr_matrix(np.random.rand(n, n))
        if 'pos' in kind:
            jac = jac.dot(jac.T)       # positively definite
            jac /= jac.max()
        if 'sym' in kind:
            jac = (jac + jac.T)/2.          # symmetric

        jac0 = jac.copy()
        rhs0 = rhs.copy()
        ijac0 = np.linalg.inv(jac.toarray())
        lhs0 = ijac0.dot(rhs0)

        self.jac = jac
        self.rhs = rhs

        # solution
        self.jac0 = jac0
        self.ijac0 = ijac0
        self.lhs0 = lhs0
        self.rhs0 = rhs0


# Dummy model with bcs
class Mbcs:
    info = "MODEL 2 - with bcs, no lm"

    def __init__(self, kind=''):
        n = 10
        rhs = np.random.rand(n)
        jac = sparse.csr_matrix(np.random.rand(n, n))
        if 'pos' in kind:
            jac = jac.dot(jac.T)       # positively definite
            jac /= jac.max()
        if 'sym' in kind:
            jac = (jac + jac.T)/2.          # symmetric

        lhs = np.zeros_like(rhs)
        bcs = np.zeros_like(rhs, dtype=bool)
        bcs[[0, -1]] = True
        lhs[bcs] = [0, 1.]

        # Solve
        jac0 = jac.copy()
        jac0[bcs] = 0.
        jac0[bcs, bcs] = 1.
        rhs0 = rhs.copy()
        rhs0[bcs] = lhs[bcs]

        ijac0 = np.linalg.inv(jac0.toarray())
        lhs0 = ijac0.dot(rhs0)

        self.rhs = rhs
        self.jac = jac
        self.lhs = lhs
        self.bcs = bcs

        # solution
        self.jac0 = jac0
        self.ijac0 = ijac0
        self.lhs0 = lhs0
        self.rhs0 = rhs0


# Dummy model with equality bcs
class Mbcseq:
    info = "MODEL 2a - with bcseq, no lm"

    def __init__(self, kind=''):
        n = 10
        rhs = np.random.rand(n)
        jac = sparse.csr_matrix(np.random.rand(n, n))
        if 'pos' in kind:
            jac = jac.dot(jac.T)       # positively definite
            jac /= jac.max()
        if 'sym' in kind:
            jac = (jac + jac.T)/2.          # symmetric

        lhs = np.zeros_like(rhs)
        bcseq = [[0, -1, [1., -2., 0.2]]]

        # preprocess and solve
        jac0 = jac.copy()
        lhs0 = lhs.copy()
        rhs0 = rhs.copy()
        i0 = bcseq[0][0]
        i1 = bcseq[0][1]
        c0, c1, c2 = bcseq[0][2]
        # modify jacobian
        jac0[i0] += jac0[i1]
        jac0[i1] = 0.
        jac0[i1, i0] = c0/c1
        jac0[i1, i1] = 1.
        # modify rhs
        rhs0[i0] += rhs0[i1]
        rhs0[i1] = c2/c1
        # solve
        ijac0 = np.linalg.inv(jac0.toarray())
        lhs0 = ijac0.dot(rhs0)
        assert np.allclose(c0*lhs0[i0] + c1*lhs0[i1], c2)

        self.rhs = rhs
        self.jac = jac
        self.lhs = lhs
        self.bcseq = bcseq

        # solution
        self.jac0 = jac0
        self.ijac0 = ijac0
        self.lhs0 = lhs0
        self.rhs0 = rhs0


# Dummy model with lm
class Mlm:
    info = "MODEL 3 - with bcs, with lm"

    def __init__(self, kind=''):
        n = 10
        nlm = 3
        rhs = np.random.rand(n)
        jac = sparse.csr_matrix(np.random.rand(n, n))
        if 'pos' in kind:
            jac = jac.dot(jac.T)       # positively definite
            jac /= jac.max()
        if 'sym' in kind:
            jac = (jac + jac.T)/2.          # symmetric

        lhs = np.zeros_like(rhs)
        bcs = np.zeros_like(rhs, dtype=bool)
        bcs[[0, -1]] = True
        lhs[bcs] = [0, 1.]

        jaclm = np.random.rand(nlm, n)
        rhslm = np.random.rand(nlm)

        jac0 = jac.copy()
        jac0[bcs] = 0.
        jac0[bcs, bcs] = 1.
        jaclm0 = jaclm.copy()
        jaclm0T = jaclm0.T
        jaclm0T[bcs] = 0.

        jac0 = sparse.bmat([[jac0, jaclm0T], [jaclm0, None]]).tocsc()

        rhs0 = rhs.copy()
        rhs0[bcs] = lhs[bcs]
        rhs0 = np.append(rhs0, rhslm)

        ijac0 = np.linalg.inv(jac0.toarray())
        lhs0 = ijac0.dot(rhs0)

        self.rhs = rhs
        self.jac = jac
        self.lhs = lhs
        self.bcs = bcs
        self.jaclm = jaclm
        self.rhslm = rhslm

        # solution
        self.jac0 = jac0
        self.ijac0 = ijac0
        self.lhs0 = lhs0
        self.rhs0 = rhs0


class Mempty:
    info = "MODEL 4 - empty"


class Mrhs:
    info = "MODEL 5 - rhs only"

    n = 10
    rhs = np.random.rand(n)


class Mlop(Mlm):
    info = "MODEL 6 - lop only"

    def __init__(self, kind=''):
        Mlm.__init__(self)
        self.jac_for_lop = self.jac
        del self.jac

    @property
    def lop(self):
        return self.jac_for_lop * self.lhs


class Mclass:
    info = "MODEL 7 - class definition"

    n = 10
    rhs = np.random.rand(n)
    lhs = np.zeros(n)
    jac = sparse.csr_matrix(np.random.rand(n, n))
    # lop = property(lambda x: jac1*lhs)


class Mfield:
    info = "MODEL 8 - model with fempy fields"

    def __init__(self, Field, ndime):
        import fempy.testing
        from common import Simple3D
        domain = eval('Simple3D().domain{}D'.format(ndime))
        # domain = Simple2D().domain2D
        rhs = Field(domain, 1.)
        n = rhs.size
        jac = sparse.csr_matrix(np.random.rand(n, n))
        # jac = np.random.rand(n, n)  # TODO: make dense to work

        self.rhs = rhs
        self.jac = jac


class Solver_Test:
    """
    This class will test all linear and nonlinear solvers, except the
    blacklisted, against linear models - non symmetric, non positive definite
    """
    def solver_test(self, solver, model, kind='', n=1, **opts):
        import time, inspect
        T = 0.
        for i in range(n):
            m = model
            if not inspect.isclass(m):  # isinstance(model, (types.TypeType, types.ClassType)):
                m = model.__class__  # Always initialize model
            try:
                m = m(kind)
            except:
                m = m()
            ls = solver(m)

            if ls.method.direct and m.jac is None:
                T = 10000000000.
                break

            if i == 0:
                # Print message
                mname = m.__class__.__name__
                nname = ls.__class__.__name__
                lname = ls.method.__class__.__name__
                pname = ls.preconditioner.__class__.__name__ \
                    if ls.preconditioner is not None else 'none'
                print("    Testing %s model against %s with %s method and"
                      " %s preconditioner" % (mname, nname, lname, pname))
                import sys
                sys.stdout.flush()

            t0 = time.time()
            ls.solve(**opts)
            T += time.time() - t0

            # Test solution
            if hasattr(m, 'lhs0'):
                x = ls.x
                # print(x, m.lhs0)
                # Test if solution is as expected
                assert np.allclose(x, m.lhs0)
            if hasattr(m, 'rhs0'):  # does not work with current ode solvers
                ls._update_P()
                ls._update_Mlm()
                x = ls.x
                b = ls.b
                J = ls.J
                M = ls.M
                # Test if jacobian is OK (jacobian is assumed to be exact)
                assert np.allclose(J.dot(x), m.rhs0)
                # Test if preconditioner is OK (only for exact precs)
                p = ls.preconditioner
                if p is not None and p.exact:
                    assert np.allclose(M.dot(b), m.lhs0)
        return T, 0

    def iterate_solvers(self, models, solvers, methods, preconditioners, N=1,
                        **options):
        methods0 = methods
        preconditioners0 = preconditioners
        for model in models:
            name = model.__name__
            if hasattr(model, 'info'):
                print(model.info)
            Tl = 100000000.
            bs = ''
            bm = ''
            bp = ''
            for solver in solvers:
                sname = solver.__name__
                if solver in blacklist:
                    continue
                if hasattr(model, 'blacklist') and sname in model.blacklist:
                    continue
                if len(methods0) == 0:
                    methods = [solver.default_method]
                for method in methods:
                    mname = method.__name__
                    solver.default_method = method
                    if len(preconditioners0) == 0:
                        preconditioners = [solver.default_preconditioner]
                    for p in preconditioners:
                        pname = p.__name__ if p is not None else 'none'
                        # skip direct methods when preconditioner is inexact
                        # print(name, sname, mname, pname)
                        if solver.default_method.direct:
                            if p is None or not p.exact:
                                continue
                        # skip blacklisted preconditioners
                        if solver in pblacklist and p in pblacklist[solver]:
                            continue
                        if hasattr(model, 'pblacklist') and \
                                mname in model.pblacklist and \
                                (len(model.pblacklist[mname]) == 0 or
                                 pname in model.pblacklist[mname]):
                            continue
                        solver.default_preconditioner = p
                        # print(name, sname, mname, pname)
                        tl, err = self.solver_test(solver, model,
                                                   kind='', n=N,
                                                   **options
                                                   )
                        if tl < Tl and not err:
                            Tl = tl
                            bs = sname
                            bm = mname
                            bp = pname
            print('Best time for model %s: %s s' % (name, Tl))
            print('Best solver: %s with %s method and %s preconditioner'
                  % (bs, bm, bp))


class Test_Linear_Solvers(Solver_Test):
    def test_mbcs(self):
        m = Mbcs()
        s = linear.BicgstabSolver(m)
        s.preconditioner = precon.lu
        s.solve()
        assert np.allclose(s.x, m.lhs0)

    def test_mlm(self):
        m = Mlm()
        s = linear.IterativeSolver(m, preconditioner=None)
        s.solve()
        assert np.allclose(s.x, m.lhs0)

    def test_linear_solvers(self, capsys):
        N = 1  # NUMBER OF SOLVER PASSAGES FOR EVERY MODEL
        models = [M, Mbcs, Mbcseq, Mlm]
        with capsys.disabled():
            print("\n")

            # First test Linear Solver against all available methods
            # and precons
            mc, pc, lsc = get_solvers()
            solvers = [linear.LinearSolver]
            methods = mc
            preconditioners = pc + [None]
            self.iterate_solvers(models, solvers, methods, preconditioners, N,
                                 tol=1e-8)

            # And then test the rest with default mthods and preconditioners
            mc, pc, lsc = get_solvers()
            solvers = lsc
            methods = []
            preconditioners = []
            self.iterate_solvers(models, solvers, methods, preconditioners, N,
                                 tol=1e-8)

    def test_direct_J_M(self):
        for model in [M, Mbcs, Mbcseq, Mlm]:
            m = linear.DirectSolver(model)
            m.update()
            b = np.random.rand(m.nlhs+m.nlm)
            assert np.allclose(m.J*(m.M*b), b)
            assert np.allclose(m.Jsparse*(m.M*b), b)

    def test_direct_J_Jsparse(self):
        for model in [M, Mbcs, Mbcseq, Mlm]:
            m = linear.DirectSolver(model)
            m.update()
            x = np.random.rand(m.nlhs+m.nlm)
            assert np.allclose(m.J*x, m.Jsparse*x)

    def test__bcseq_unroll(self):
        models = [Mbcs()]  # , Mbcseq, Mlm]
        for m in models:
            m.bcseq = [[0, 1], [-2, -1], [2, 3], [1, 0], [4, 4], [0, -1],
                       [2, 1], [1, 4]]
            bcseq0, bcseq1, bcseqc = linear._unroll_bcseq(m)
            assert np.allclose(bcseq0, [0, -1, 2, 1])
            assert np.allclose(bcseq1, [1, -2, 3, 4])
            assert np.allclose(bcseqc.T[0], [0., -1., 0])
            assert np.allclose(bcseqc.T[1], [0., 1., 1.])
            assert np.allclose(bcseqc.T[2], [1., -1., 0.])
            assert np.allclose(bcseqc.T[3], [1., -1., 0.])


class Test_Basic(Solver_Test):
    def test_empty_model(self):
        with pytest.raises(AttributeError):
            linear.LinearSolver(Mempty)

    def test_rhs_only_model(self):
        with pytest.raises(AttributeError):
            linear.LinearSolver(Mrhs).solve()

    def test_lop_only_model(self):
        linear.IterativeSolver(Mlop()).solve()

    def test_class_model(self):
        linear.LinearSolver(Mclass).solve()


class Test_Field_Solve(Solver_Test):
    def test_field_solve(self):
        from fempy.fields import nfields
        self.field_solve(nfields.nScalar, ndime=1)
        self.field_solve(nfields.nScalar, ndime=2)
        self.field_solve(nfields.nScalar, ndime=3)

        self.field_solve(nfields.nVector, ndime=1)
        self.field_solve(nfields.nVector, ndime=2)
        self.field_solve(nfields.nVector, ndime=3)

        self.field_solve(nfields.nTensor, ndime=1)
        self.field_solve(nfields.nTensor, ndime=2)
        self.field_solve(nfields.nTensor, ndime=3)

#        self.field_solve(nfields.nTensor3, ndime=1)
#        self.field_solve(nfields.nTensor3, ndime=2)
#        self.field_solve(nfields.nTensor3, ndime=3)
#
#        self.field_solve(nfields.nTensor4, ndime=1)
#        self.field_solve(nfields.nTensor4, ndime=2)
#        self.field_solve(nfields.nTensor4, ndime=3)

    def field_solve(self, Field, ndime):
        Solver = linear.DirectSolver

        # no bcs
        m = Mfield(Field, ndime)
        ls = Solver(m)
        ls.solve()
        # assert type(xn) is Field

        r = ls.residual()
        A = m.jac
        x = m.lhs.ravel()
        b = m.rhs.ravel()
        assert np.allclose(r, A.dot(x) - b)

        # with bcs
        m = Mfield(Field, ndime)
        m.bcs = np.zeros_like(m.rhs, dtype=bool)
        m.bcs[[0, -1]] = True
        ls = Solver(m)
        ls.solve()
        # assert type(xn) is Field

        r = m.res.ravel()
        A = m.jac
        x = m.lhs.ravel()
        b = m.rhs.ravel()
        # b at fixed nodes should remain???
        assert np.allclose(r, A.dot(x) - b)

#        # with bcs - all True
#        m = Mfield(Field, ndime, nnode)
#        m.bcs = np.zeros_like(m.rhs, dtype=bool)
#        m.bcs[:] = True
#        ls = Solver(m)
#        xn = ls.solve()
#        assert type(xn) is Field
#
#        r = ls.residual()
#        A = m.jac
#        x = m.lhs.ravel()
#        b = m.rhs.ravel()
#        # b at fixed nodes should remain???
#        assert np.allclose(r, A.dot(x) - b)

        # with lm
        m = Mfield(Field, ndime)
        m.bcs = np.zeros_like(m.rhs, dtype=bool)
        m.bcs[[0, -1]] = True
        m.jaclm = np.random.rand(3, m.bcs.size)
        ls = Solver(m)
        ls.solve()
        # assert type(xn) is Field

        r = np.concatenate([ri.ravel() for ri in m.res])
        A = m.jac
        x = m.lhs.ravel()
        b = m.rhs.ravel()
        B = m.jaclm
        lm = m.lhslm
        c = m.rhslm
        rx = A.dot(x) + B.T.dot(lm) - b
        rlm = B.dot(x) - c
        rall = np.concatenate((rx, rlm))
        assert np.allclose(r, rall)


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 27 21:30:46 2014

@author: mwojc
"""

import inspect
import fempy.solvers.nonlin as nonlin
from fempy.domain.tools import cached_property


class Model(object):
    default_solver = nonlin.NewtonKrylovSolver

    def __init__(self, *args, **kwargs):
        """
        Model initialization sequence:
        1. Look for domain in args (first argument) or kwargs
        2. If domain is not found: try to call domain_setup which should
           set self.domain atribute, It is called *without* arguments
        2. If domain is found just set self.domain
        3. Call init - init is called *without* arguments!
        4. Look for solver in kwargs
        5. If solver not found set self.solver = self.default_solver
        6. If solver is found set it as self.solver
        7. Finally call self.setup method with unmodified *args and **kwargs
        """
        # Try to discover domain
        domain = kwargs.pop('domain', None)
        if len(args) and hasattr(args[0], 'nnode'):
            domain = args[0]
            args = args[1:]
            kwargs.pop('domain', None)  # remove domain from kwargs
        if domain is None:
            self.setup_domain()
        else:
            self.domain = domain
        if hasattr(self, 'init'):
            self.init(**kwargs)
        # now set solver...
        if 'solver' in kwargs:
            solver = kwargs['solver']
            del kwargs['solver']
        else:
            solver = self.default_solver
        self.solver = solver
        # and make setups
        if hasattr(self, 'setup'):
            self.setup(*args, **kwargs)
        self.setup_model(*args, **kwargs)

    def setup_domain(self, *args, **kwargs):
        raise TypeError("Provide a domain or define 'setup_domain' method")

    def setup_model(self, *args, **kwargs):
        pass

    def update(self):
        # get all names in model class and object
        c = self.__class__
        d = self.__dict__
        cdir = dir(c)
        dkeys = d.keys()
        for n in cdir:
            # find properties and invalidate them if they have been 'cached'
            if isinstance(getattr(c, n), cached_property) and n in dkeys:
                # property has been cached so we invalidate it
                del d[n]
            # now find all update functions and call them (without arguments)
            if (n.endswith('_update') or
                    (n.endswith('_') and not n.endswith('__'))):
                attr = getattr(self, n)
                if callable(attr):
                    attr()

    def invalidate_caches(self, cdir=None, dkeys=None):
        # get all names in model class and object
        c = self.__class__
        d = self.__dict__
        cdir = dir(c)
        dkeys = d.keys()
        for n in cdir:
            if isinstance(getattr(c, n), cached_property) and n in dkeys:
                del d[n]

    def solve(self, **opts):
        self.solver.solve(**opts)
        return self.lhs

    def preview(self, what='lhs', time=()):
        from fempy.geometry import gmsh_io
        gmsh_io.preview_model(self, what, time)

    def preview_dump(self, fname, what='lhs', time=(), ascii=False):
        from fempy.geometry import gmsh_io
        gmsh_io.dump_model(fname, self, what, time, ascii)

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, s):
        if inspect.isclass(s):
            self._solver = s(self)
        else:
            self._solver = s
            # TODO: check if 's' is proper solver for the model


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 19 07:14:03 2014

@author: mwojc
"""
import numpy as np
from fempy.geometry import geometry as g


def _group_prefix(group):
    if group == '':
        prefix = ''
        group = 'domain'
    else:
        prefix = str(group) + '_'
        group = group
    return group, prefix


def Rectangle(a=1., b=1., lb=None, group='', periodic=False, **kwargs):
    if lb is None:
        lb = -a/2., -b/2.
    group, prefix = _group_prefix(group)
    x0, y0 = lb
    A = g.Point([x0, y0],
                groups=[prefix + 'bottom_left', prefix + 'corners'])
    B = g.Point([x0+a, y0],
                groups=[prefix + 'bottom_right', prefix + 'corners'])
    C = g.Point([x0+a, y0+b],
                groups=[prefix + 'top_right', prefix + 'corners'])
    D = g.Point([x0, y0+b],
                groups=[prefix + 'top_left', prefix + 'corners'])

    AB = g.Line(A, B, [prefix + 'bottom', prefix + 'boundary'])
    BC = g.Line(B, C, [prefix + 'right', prefix + 'boundary'])
    CD = g.Line(C, D, [prefix + 'top', prefix + 'boundary'])
    DA = g.Line(D, A, [prefix + 'left', prefix + 'boundary'])

    if periodic:
        AB.periodic(CD)
        DA.periodic(BC)

    ABCD = g.PlaneSurface([AB, BC, CD, DA], group, **kwargs)
    if 'elsize' not in kwargs:
        ABCD.elsize = min([a, b])/8.
    return ABCD


class Shape(object):
    pass


class Rectangle2(Shape):
    """
    D               C
    +---------------+
    |               |
    |               |
    |       +       |
    |     (0,0)     |
    |               |
    +---------------+
    A               B
    """
    def __init__(self, a=1., b=1., lb=None, elsize=None, periodic=False):
        if lb is None:
            lb = -a/2., -b/2.
        x0, y0 = lb
        A = g.Point([x0, y0], groups=['A', 'bottom_left', 'corners'])
        B = g.Point([x0+a, y0], groups=['B', 'bottom_right', 'corners'])
        C = g.Point([x0+a, y0+b], groups=['C', 'top_right', 'corners'])
        D = g.Point([x0, y0+b], groups=['D', 'top_left', 'corners'])

        AB = g.Line(A, B, ['AB', 'bottom'])
        BC = g.Line(B, C, ['BC', 'right'])
        CD = g.Line(C, D, ['CD', 'top'])
        DA = g.Line(D, A, ['DA', 'left'])

        if elsize is None:
            elsize = min([a, b])/8.

        if periodic:
            AB.periodic(CD)
            DA.periodic(BC)

        ABCD = g.PlaneSurface([AB, BC, CD, DA], elsize=elsize)

        self.A = A
        self.B = B
        self.C = C
        self.D = D
        self.AB = AB
        self.BC = BC
        self.CD = CD
        self.DA = DA
        self.ABCD = ABCD

    def domain(self, order=1, verbose=False):
        d = self.ABCD.gmsh.domain(order, verbose)
        d.geometry = self
        return d

    def preview(self, default='geomsh'):
        self.ABCD.gmsh.preview(default)


def Rectangles(a=[1., 0.5], b=[1., 0.5], lb=None, hole=False, **kwargs):
    assert len(a) == len(b)     # a and b have to be of the same length
    assert a == sorted(a, reverse=True)  # a have to be in descending order
    assert b == sorted(b, reverse=True)  # b have to be in descending order
    rectangles = [Rectangle(a[0], b[0], group='D0', **kwargs)]
    for i in range(1, len(a)):
        rec = Rectangle(a[i], b[i], group='D'+str(i), **kwargs)
        rectangles[-1].add_hole(rec)
        rectangles.append(rec)
    if hole is True:
        rectangles = rectangles[:-1]
    G = g.Geometry(rectangles)
    if lb is not None:
        vec = (lb[0]+a[0]/2., lb[1]+b[0]/2.)
        G.move(vec)

    if len(rectangles) == 1:
        return rectangles[0]  # if single circle with hole
    else:
        return G


def Polygon(coors, group='', close=True, surface=True, **kwargs):
    coors = g.tools.coors3D(coors)
    group, prefix = _group_prefix(group)
    pts = [g.Point(c) for c in coors]
    is_closed = False
    if pts[0].eq(pts[-1]):
        pts[-1] = pts[0]
        is_closed = True
    if close:
        if pts[0].ne(pts[-1]):
            pts += [pts[0]]  # Close polygon if not closed
            is_closed = True
    lines = []
    for i in range(len(pts)-1):
        l = g.Line(pts[i], pts[i+1], [prefix+'line%i' % i,
                                      prefix+'boundary'])
        lines += [l]
    if is_closed and surface:
        srf = g.PlaneSurface(lines, group, **kwargs)
        return srf
    else:
        geo = g.Geometry(lines, group=group, **kwargs)
        return geo


def Layers2D(H=[1.], L=[1.], firstL=True, elsize=None, elsize_factor=1.):
    points = []
    lines = []
    left = []
    right = []
    layers = []
    if elsize is None:
        def elsize(x, y):
            He = np.array(H)
            He = np.abs(He)
            He = He[He > 1e-15]
            return min(min(He)/3., sum(He)/20.)
    if isinstance(elsize, (int, float)):
        def elsize(x, y):
            return float(elsize)
    for i, h in enumerate(H):
        if i > 0 and firstL and len(L) > 2:
            L = [L[0], L[-1]]
        pts = []
        for j, l in enumerate(L):
            els = elsize(l, h) * elsize_factor
            pts.append(g.Point([l, h], groups='p%s%s' % (i, j), elsize=els))
        lin = [g.Line(p0, p1) for p0, p1 in zip(pts[:-1], pts[1:])]
        for k, l in enumerate(lin):
            l.add_to_groups(['line%s%s' % (i, k), 'line%s' % i])
            if i == 0:
                l.add_to_groups('top')
            if i == len(H)-1:
                l.add_to_groups('bottom')
        if i > 0:
            pts0 = points[i-1]
            lin0 = lines[i-1]
            llin = g.Line(pts0[0], pts[0], groups='left')
            rlin = g.Line(pts0[-1], pts[-1], groups='right')
            layer = g.PlaneSurface(lin + [rlin] + lin0 + [llin],
                                   groups='layer%s' % str(i-1))
            left.append(llin)
            right.append(rlin)
            layers.append(layer)
        points.append(pts)
        lines.append(lin)
#    if elsize is None:
#        H = np.array(H)
#        H = np.abs(H)
#        H = H[H > 1e-15]
#        elsize = min(min(H)/3., sum(H)/20.)*elsize_factor
    geo = g.Geometry(layers) #, elsize=elsize)
    geo.layers = layers
    geo.lines = lines
    geo.right = right
    return geo


def Circle(r=1., center=(0., 0.), group='', **kwargs):
    xc, yc = center
    group, prefix = _group_prefix(group)  # default group is 'domain'
    O = g.Point(center)
    A = g.Point((xc+r, yc), groups=[prefix+'A'])
    B = g.Point((xc, yc+r), groups=[prefix+'B'])
    C = g.Point((xc-r, yc), groups=[prefix+'C'])
    D = g.Point((xc, yc-r), groups=[prefix+'D'])
    AOB = g.CircularArc(A, O, B, [prefix+'q1', prefix+'boundary'])
    BOC = g.CircularArc(B, O, C, [prefix+'q2', prefix+'boundary'])
    COD = g.CircularArc(C, O, D, [prefix+'q3', prefix+'boundary'])
    DOA = g.CircularArc(D, O, A, [prefix+'q4', prefix+'boundary'])
    if 'elsize' not in kwargs:
        kwargs['elsize'] = r/5.
    AOA = g.PlaneSurface([AOB, BOC, COD, DOA], group, **kwargs)
    return AOA


def Circles(r=[1.0, 0.5], center=(0, 0), hole=False, **kwargs):
    assert r == sorted(r, reverse=True)  # r have to be in descending order
    circles = [Circle(r[0], center=center, group='D0', **kwargs)]
    for i in range(1, len(r)):
        circ = Circle(r[i], center=center, group='D'+str(i), **kwargs)
        circles[-1].add_hole(circ)
        circles.append(circ)
    if hole is True:
        circles = circles[:-1]

    if len(circles) == 1:
        return circles[0]  # if single circle with hole
    else:
        return g.Geometry(circles)


def Ellipse(r=(1., 0.5), center=(0, 0), group='', **kwargs):
    ra, rb = r
    xc, yc = center
    group, prefix = _group_prefix(group)  # default group is 'domain'
    O = g.Point(center)
    A = g.Point((xc+ra, yc))
    B = g.Point((xc, yc+rb))
    C = g.Point((xc-ra, yc))
    D = g.Point((xc, yc-rb))
    AOAB = g.EllipticalArc(A, O, A, B, [prefix+'q1', prefix+'boundary'])
    BOAC = g.EllipticalArc(B, O, A, C, [prefix+'q2', prefix+'boundary'])
    COAD = g.EllipticalArc(C, O, A, D, [prefix+'q3', prefix+'boundary'])
    DOAA = g.EllipticalArc(D, O, A, A, [prefix+'q4', prefix+'boundary'])
    if 'elsize' not in kwargs:
        kwargs['elsize'] = sum(r)/10.
    AOA = g.PlaneSurface([AOAB, BOAC, COAD, DOAA], group, **kwargs)
    return AOA


def Ellipses(r=[(1.0, 0.5), (0.5, 0.25)], center=(0, 0), hole=False, **kwargs):
    ra, rb = zip(*r)
    assert list(ra) == sorted(ra, reverse=True)  # in descending order?
    assert list(rb) == sorted(rb, reverse=True)  # in descending order?
    elps = [Ellipse(r=(ra[0], rb[0]), center=center, group='D0', **kwargs)]
    for i in range(1, len(r)):
        ei = Ellipse(r=(ra[i], rb[i]), center=center,
                     group='D'+str(i), **kwargs)
        elps[-1].add_hole(ei)
        elps.append(ei)
    if hole is True:
        elps = elps[:-1]

    if len(elps) == 1:
        return elps[0]  # if single circle with hole
    else:
        return g.Geometry(elps)


def RandomHull(r=1., n=10, center=(0, 0), res=200, clcurve=0, group='',
               **kwargs):
    assert n > 2

    from scipy.spatial import ConvexHull
    from shapely.geometry import LinearRing

    group, prefix = _group_prefix(group)  # default group is 'domain'

    def hull():
        R = np.random.uniform(0, r, n)
        theta = np.random.uniform(0, 2*np.pi, n)
        x0, y0 = center
        x = x0 + R * np.cos(theta)
        y = y0 + R * np.sin(theta)
        coors = np.asarray([x, y]).T
        hull = ConvexHull(coors)
        hcoors = coors[hull.vertices]
        nh = 2
        dist = r/4.
        while nh <= 2:
            rhcoors = g.tools.reduce_coors(hcoors, dist)
            nh = len(rhcoors)
            dist = dist/2.
        points = [g.Point(hc) for hc in rhcoors]
        points = points + [points[0]]
        return points

    spline = g.Spline(hull(), groups=prefix+'boundary')
    vertices = spline.poly(res=res)
    while not LinearRing(vertices).is_simple:
        spline = g.Spline(hull(), groups=prefix+'boundary')
        vertices = spline.poly(res=res)
    if 'elsize' not in kwargs:
        radii = np.linalg.norm((vertices - vertices.mean(axis=0)), axis=1)
        elsize = radii.mean()/3.
        kwargs['elsize'] = elsize
    surface = g.PlaneSurface([spline], groups=group, **kwargs)
    surface.gmsh.options.Mesh.CharacteristicLengthFromCurvature = clcurve
    return surface


def FlowRVE(r=1., h=0.1, elsize_factor=1., full=False):
    #ri = -0.585786437626905*h + 0.414213562373095*r
    h = h/2.
    ri = np.sqrt(2)*(r+h)-r-2*h
    mid = (h+r) - r*np.sqrt(2)/2.

    p0 = g.Point([-r-h, -h])
    p1 = g.Point([-h, -r-h])
    p2 = g.Point([h, -r-h])
    p3 = g.Point([r+h, -h])
    p4 = g.Point([r+h, h])
    p5 = g.Point([h, r+h])
    p6 = g.Point([-h, r+h])
    p7 = g.Point([-r-h, h])
    p8 = g.Point([-r-h, -r-h])
    p9 = g.Point([r+h, -r-h])
    p10 = g.Point([r+h, +r+h])
    p11 = g.Point([-r-h, +r+h])

    p6x = g.Point([-mid, mid], elsize=r/10.)
    p0x = g.Point([-mid, -mid], elsize=r/10.)
    p2x = g.Point([mid, -mid], elsize=r/10.)
    p4x = g.Point([mid, mid], elsize=r/10.)

    c0 = g.CircularArc(p6, p11, p6x, 'outer_boundary, fixed, upper_left')
    c0x = g.CircularArc(p6x, p11, p7, 'outer_boundary, fixed, upper_left')
    c1 = g.CircularArc(p0, p8, p0x, 'outer_boundary, fixed, lower_left')
    c1x = g.CircularArc(p0x, p8, p1, 'outer_boundary, fixed, lower_left')
    c2 = g.CircularArc(p2, p9, p2x, 'outer_boundary, fixed, lower_right')
    c2x = g.CircularArc(p2x, p9, p3, 'outer_boundary, fixed, lower_right')
    c3 = g.CircularArc(p4, p10, p4x, 'outer_boundary, fixed, upper_right')
    c3x = g.CircularArc(p4x, p10, p5, 'outer_boundary, fixed, upper_right')
    c4 = g.Line(p5, p6, 'outer_boundary, flow, top', elsize=h/3.)
    c5 = g.Line(p7, p0, 'outer_boundary, flow, left', elsize=h/3.)
    c6 = g.Line(p1, p2, 'outer_boundary, flow, bottom', elsize=h/3.)
    c7 = g.Line(p3, p4, 'outer_boundary, flow, right', elsize=h/3.)

    srf = g.PlaneSurface([c0, c0x, c5, c1, c1x, c6, c2, c2x, c7, c3, c3x, c4],
                         'domain, matrix_domain')

    hole = Circle(ri, group='inner')
    srf.add_hole(hole)
    srf.elsize_factor = elsize_factor

    if not full:
        return srf

    l0 = g.Line(p6, p11, 'outer_boundary, fixed, top')
    l0x = g.Line(p11, p7, 'outer_boundary, fixed, left')
    srf1 = g.PlaneSurface([l0, l0x, c0, c0x], 'outer_void')
    l1 = g.Line(p0, p8, 'outer_boundary, fixed, left')
    l1x = g.Line(p8, p1, 'outer_boundary, fixed, bottom')
    srf2 = g.PlaneSurface([l1, l1x, c1, c1x], 'outer_void')
    l2 = g.Line(p2, p9, 'outer_boundary, fixed, bottom')
    l2x = g.Line(p9, p3, 'outer_boundary, fixed, right')
    srf3 = g.PlaneSurface([l2, l2x, c2, c2x], 'outer_void')
    l3 = g.Line(p4, p10, 'outer_boundary, fixed, right')
    l3x = g.Line(p10, p5, 'outer_boundary, fixed, top')
    srf4 = g.PlaneSurface([l3, l3x, c3, c3x], 'outer_void')
    return srf+srf1+srf2+srf3+srf4


def Box(a=1., b=1., c=1., origin=(0., 0., 0.), elsize=None, elsize_factor=1.,
        group='', periodic=False, **kwargs):
    from fempy.geometry import Gmsh
    elsize = min((a, b, c))/5. if elsize is None else elsize
    elsize *= elsize_factor
    code = \
"""
SetFactory("OpenCASCADE");
v0 = newv;
Box(v0) = {{{}}};
pts_v0[] = PointsOf{{Volume{{v0}};}};
Characteristic Length{{pts_v0[]}} = {};
""".format(str(origin+(a, b, c))[1:-1], elsize)
    geometry = Gmsh(code, entities={3: {1: 'domain'},
                                    2: {1: 'back', 2: 'front',
                                        3: 'left', 4: 'right',
                                        5: 'bottom', 6: 'top'},
                                    0: {2: 'origin', 7: 'abc'}
                                    })
    return geometry


def Cylinder(x0=(0, 0, 0.), axis=(0, 0, 1.), radius=0.5, angle=None,
             elsize=None, elsize_factor=1., **kwargs):
    import pygmsh
    from fempy.geometry import Gmsh
    elsize = 0.2 * radius * elsize_factor if elsize is None else elsize
    g = pygmsh.opencascade.Geometry()
    g.add_cylinder(x0, axis, radius, angle=angle, char_length=elsize)
    geometry = Gmsh(g, entities={3: {1: 'domain'},
                                 2: {1: 'side', 2: 'top', 3: 'bottom'},
                                 1: {1: 'top_egde', 2: 'side_edge',
                                     3: 'bottom_egde'},
                                 0: {1: 'top_point', 2: 'bottom_point'},
                                 })
    return geometry


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')

#    g = FlowRVE(full=True)
#    g.gmsh.preview()

#    def rigid_bcs(d):
#        # Discover points for fixing rigid body movement
#        b = d.boundary
#        bcoors = b.coors
#        tol = abs(bcoors.max()-bcoors.min()) * 1e-14
#        X, Y, Z = bcoors.T
#        # first point - for all dofs fixing
#        idx = np.nonzero(np.abs(X - X.min()) < tol)[0]
#        if len(idx) > 1:
#            idx = idx[np.abs(Y[idx] - Y[idx].min()) < tol]
#        if len(idx) > 1:
#            idx = idx[np.abs(Z[idx] - Z[idx].min()) < tol]
#        b.ngroups['fix0'] = idx
#        # second point for fixing second dof
#        idx = np.nonzero(np.abs(Y - Y.min()) < tol)[0]
#        if len(idx) > 1:
#            idx = idx[np.abs(X[idx] - X[idx].max()) < tol]
#        if len(idx) > 1:
#            idx = idx[np.abs(Z[idx] - Z[idx].min()) < tol]
#        b.ngroups['fix1'] = idx
#        # other points are necessary only for 3D
#        if d.ndime == 3:
#            # third point for fixing third dof
#            idx = np.nonzero(np.abs(Y - Y.max()) < tol)[0]
#            if len(idx) > 1:
#                idx = idx[np.abs(X[idx] - X[idx].min()) < tol]
#            if len(idx) > 1:
#                idx = idx[np.abs(Z[idx] - Z[idx].min()) < tol]
#            b.ngroups['fix2'] = idx
#            # fourth point for fixing first dof
#            idx = np.nonzero(np.abs(Z - Z.max()) < tol)[0]
#            if len(idx) > 1:
#                idx = idx[np.abs(X[idx] - X[idx].max()) < tol]
#            if len(idx) > 1:
#                idx = idx[np.abs(Y[idx] - Y[idx].max()) < tol]
#            b.ngroups['fix3'] = idx
#
#
#    from fempy.models.elastic import LinearElastic3D, Axisymmetry
#    from fempy.geometry.gmsh_io import preview_field
#
#    class Thermal3D(LinearElastic3D):
#        def setup_domain(self):
#            self.geometry = Cylinder(elsize=0.2)
#            self.domain = self.geometry.gmsh.domain()
#
#        def setup_model(self):
#            rigid_bcs(self.domain)
#            self.bcs['fix0'] = [True, True, True]
#            self.bcs['fix1'] = [False, True, False]
#            self.bcs['fix2'] = [False, False, True]
#            self.bcs['fix3'] = [True, False, False]
#            self.ev(20000., 0.3)
#            self.alpha(1e-4)
#            self.T[:] += 3.
#
#    m3D = Thermal3D()
#    m3D.solve()
#    preview_field(m3D.sigma[..., 1, 1], name='sigma33', d=m3D.domain)
#
#
#    class Thermal2DPlaneStrain(PlaneStrain):
#        def setup_domain(self):
#            self.geometry = Rectangle(10, 10, lb=(0, 0), elsize=0.3)
#            self.domain = self.geometry.gmsh.domain()
#
#        def setup_model(self):
#            self.bcs['bottom_left'] = [True, True]
#            self.bcs['bottom_right'] = [False, True]
#            self.ev(20000., 0.3)
#            self.alpha[:] = 1e-4 * np.array([[1, 0], [0, 1]])
#            self.T[:] += 3.
#
#    m2D = Thermal2DPlaneStrain()
#    m2D.solve()
#    preview_field(m2D.sigmaz, name='sigma33',d=m2D.domain)



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:54:09 2016

@author: mwojc
"""
from __future__ import print_function
from fempy.compat import reload
import pytest
import inspect
import fempy.solvers.methods as methods
import fempy.solvers.precon as precon
import fempy.solvers.ivp as ivp
from solvers import test_linear


def get_solvers():
    reload(methods)
    reload(precon)
    reload(ivp)
    mc = [ls for ls in methods.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, methods.linear)]

    pc = [ls for ls in precon.__dict__.values()
          if inspect.isclass(ls) and issubclass(ls, precon.preconditioner)]

    isc = [ls for ls in ivp.__dict__.values()
           if inspect.isclass(ls) and issubclass(ls, ivp.ODESolver)]
    return mc, pc, isc


class M(test_linear.M):
    pblacklist = {'gmres': 'diag'}

    def __init__(self, kind=''):
        test_linear.M.__init__(self, kind)
        del self.rhs0
        self.rhs_tot = self.rhs.copy()
        self.rhs[:] = 0.
        self.time0 = 0.
        self.time1 = 10000000000000.

    def update(self):
        self.rhs[:] = self.time*self.rhs_tot/10000000000000.


class Mbcs(test_linear.Mbcs):
    pblacklist = {'gmres': 'diag'}

    def __init__(self, kind=''):
        test_linear.Mbcs.__init__(self, kind)
        del self.rhs0
        self.rhs_tot = self.rhs.copy()
        self.lhs_bcs_tot = self.lhs[self.bcs].copy()
        self.rhs[:] = 0.
        self.time0 = 0.
        self.time1 = 10000000000000.

    def update(self):
        self.rhs[:] = self.time*self.rhs_tot/10000000000000.
        self.lhs[self.bcs] = self.time*self.lhs_bcs_tot/10000000000000.


class Mbcseq(test_linear.Mbcseq):
    pblacklist = {'gmres': 'diag'}

    def __init__(self, kind=''):
        test_linear.Mbcseq.__init__(self, kind)
        del self.rhs0
        self.rhs_tot = self.rhs.copy()
        self.delta_bcseq_tot = self.bcseq[0][2][-1]
        self.rhs[:] = 0.
        self.time0 = 0.
        self.time1 = 10000000000000.

    def update(self):
        self.rhs[:] = self.time*self.rhs_tot/10000000000000.
        self.bcseq[0][2][-1] = self.time*self.delta_bcseq_tot/10000000000000.


class Mlm(test_linear.Mlm):
    pblacklist = {'gmres': 'diag'}

    def __init__(self, kind=''):
        test_linear.Mlm.__init__(self, kind)
        del self.rhs0
        self.rhs_tot = self.rhs.copy()
        self.lhs_bcs_tot = self.lhs[self.bcs].copy()
        self.rhslm_tot = self.rhslm.copy()
        self.rhs[:] = 0.
        self.time0 = 0.
        self.time1 = 10000000000000.

    def update(self):
        self.rhs[:] = self.time*self.rhs_tot/10000000000000.
        self.lhs[self.bcs] = self.time*self.lhs_bcs_tot/10000000000000.
        self.rhslm[:] = self.time*self.rhslm_tot/10000000000000.


class Test_Ivp_Solvers(test_linear.Solver_Test):
    def test_ivp_solvers(self, capsys):
        N = 1  # NUMBER OF SOLVER PASSAGES FOR EVERY MODEL
        models = [M, Mbcs, Mlm, Mbcseq]  # , [Mscipy]

        with capsys.disabled():
            print("\n")

            # First test solvers capable to use fempy linear methods
            # against all available methods and precons
            mc, pc, _ = get_solvers()
            solvers = [ivp.BDFSolver]
            methods = mc
            preconditioners = pc + [None]
            self.iterate_solvers(models, solvers, methods, preconditioners, N,)
                                 # verbose=True)

            # And then test the rest with default methods and preconditioners
            _, _, solvers = get_solvers()
            methods = []
            preconditioners = []
            self.iterate_solvers(models, solvers, methods, preconditioners, N)
                                 # verbose=True)

    def test_compare_BDF_bcseq(self):
        import numpy as np
        m = Mbcseq()
        s1 = ivp.BDFSolverDS(m)
        s1.update()
        s2 = ivp.BDFSolver(m)
        s2._verbose = False
        s2.update()
        s2.update_c(0.1)

        Js1 = s1.Jsparse
        I = np.eye(10)
        Js1 = I - 0.1 * s1.Jsparse
        Js2 = s2.Jsparse.toarray()

        assert np.allclose(Js1, Js2)

        x = np.random.rand(10)
        assert np.allclose(s2.J*x, s2.Jsparse*x)

    def test_compare_BDF_bcs(self):
        import numpy as np
        m = Mbcs()
        s1 = ivp.BDFSolverDS(m)
        s1.update()
        s2 = ivp.BDFSolver(m)
        s2._verbose = False
        s2.update()
        s2.update_c(0.1)

        Js1 = s1.Jsparse
        I = np.eye(10)
        Js1 = I - 0.1 * s1.Jsparse
        Js2 = s2.Jsparse.toarray()

        assert np.allclose(Js1, Js2)

        x = np.random.rand(10)
        assert np.allclose(s2.J*x, s2.Jsparse*x)

    def test_compare_BDF_lm(self):
        import numpy as np
        m = Mlm()
        s1 = ivp.BDFSolverDS(m)
        s1.update()
        s2 = ivp.BDFSolver(m)
        s2._verbose = False
        s2.update()
        s2.update_c(0.1)
        
        Js1 = s1.Jsparse
        I = np.eye(13)
        Js1 = I - 0.1 * s1.Jsparse
        Js2 = s2.Jsparse.toarray()
        
        # print(Js1 - Js2)
        assert np.allclose(Js1, Js2)
        
        x = np.random.rand(13)
        # print(s2.J*x)
        # print(s2.Jsparse*x)
        assert np.allclose(s2.J*x, s2.Jsparse*x)

    def test_compare_BDF(self):
        import numpy as np
        m = M()
        s1 = ivp.BDFSolverDS(m)
        s1.update()
        s2 = ivp.BDFSolver(m)
        s2._verbose=False
        s2.update()
        s2.update_c(0.1)
        
        Js1 = s1.Jsparse
        I = np.eye(10)
        Js1 = I - 0.1 * s1.Jsparse
        Js2 = s2.Jsparse.toarray()
        
        # print(Js1 - Js2)
        assert np.allclose(Js1, Js2)
        
        x = np.random.rand(10)
        # print(s2.J*x)
        # print(s2.Jsparse*x)
        assert np.allclose(s2.J*x, s2.Jsparse*x)

    def test_M_J_bcseq(self):
        import numpy as np
        m = ivp.BDFSolver(Mbcseq)
        m._verbose = True
        m.update()
        m.update_c(0.1)
        b = np.random.rand(10)
        # print(m.J*(m.M*b) - b)
        # print(m.Jsparse*(m.M*b))
        # print(b)
        # print(m.M*b)
        Js = m.Jsparse
        from scipy.sparse import linalg
        splu = linalg.splu(Js)
        # print(splu.solve(b))
        # print(m.M*b - splu.solve(b))
        # print Js[0]
        assert np.allclose(m.J*(m.M*b), b)
        assert np.allclose(m.Jsparse*(m.M*b), b)

    def test_M_J_bcs(self):
        import numpy as np
        m = ivp.BDFSolver(Mbcs)
        m._verbose = True
        m.update()
        m.update_c(0.1)
        b = np.random.rand(10)
        # print(m.J*(m.M*b) - b)
        # print(m.Jsparse*(m.M*b))
        # print(b)
        # print(m.M*b)
        #
        # Js = m.Jsparse
        # from scipy.sparse import linalg
        # splu = linalg.splu(Js)
        # print(splu.solve(b))
        # print(Js.toarray() - (m.I - m.c*m.K))
        assert np.allclose(m.J*(m.M*b), b)
        assert np.allclose(m.Jsparse*(m.M*b), b)

    def test_M_J_lm(self):
        import numpy as np
        m = ivp.BDFSolver(Mlm)
        m._verbose = True
        m.update()
        m.update_c(0.1)
        b = np.random.rand(13)
        # print(m.J*(m.M*b) - b)
        assert np.allclose(m.J*(m.M*b), b)
        assert np.allclose(m.Jsparse*(m.M*b), b)

    def test_M_J(self):
        import numpy as np
        m = ivp.BDFSolver(M)
        m._verbose = True
        m.update()
        m.update_c(0.1)
        b = np.random.rand(10)
        assert np.allclose(m.J*(m.M*b), b)
        assert np.allclose(m.Jsparse*(m.M*b), b)


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 11:05:58 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.nfields import nField, nScalar, nVector, nList
from fempy.fields.nfields import nTensor, nMatrix, nTensor3, nTensor4
from fempy.fields.algebra import dot, tensordot
from fempy.testing import testdir
from common import Simple3D
import pytest


class Test_nField(Simple3D):
    @property
    def domain(self):
        return self.domain3D

    @property
    def field(self):
        return nTensor(self.domain, np.random.rand(len(self.coors), 3, 3))

    def test_getitem(self):
        d = self.domain
        f = nTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        for k, v in d.ngroups.items():
            assert (v == f.groups[k]).all()
        d.ngroups['test2'] = slice(1, 6, 2)
        assert (f['test2'] == f[1:6:2]).all()
        assert np.allclose(f['tope'], f[d.eidx.nidx['tope']])

        with pytest.raises(AttributeError):
            f.groups = {}
        eg = d.ngroups
        d.ngroups = {}
        assert f.groups == {}
        d.ngroups = eg

    def test_setitem(self):
        d = self.domain
        f = nTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        f['tope'] = 5.
        assert np.allclose(f[7, 0], [5., 5., 5.])
        assert np.allclose(f[0, 0], [1., 1., 0.])
        assert np.allclose(f['tope', 0, 0], [5.]*9)

    def test_scalar_field_creation(self):
        d = self.domain
        f = nScalar(d)
        assert type(f) == nScalar
        assert f.shape == (d.nnode,)
        assert np.allclose(f, 0.)
        assert f.domain is d
        f = nScalar(d, 1.)
        assert type(f) == nScalar
        assert f.shape == (d.nnode,)
        assert np.allclose(f, 1.)
        f = nScalar(d, f)
        assert type(f) == nScalar
        with pytest.raises(AssertionError):
            f = nScalar(d, [2.])
        with pytest.raises(AssertionError):
            f = nScalar(d, [[2.], [1]])
        f = nScalar(d, 'test')
        assert f.dtype in ('S4', '<U4')

    def test_vector_field_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nVector(d)
        assert type(f) == nVector
        assert f.shape == (nn, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nVector(d, 1.)
        assert f.shape == (nn, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        f = nVector(d, [1., 1., 0.])
        assert f.shape == (nn, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = nVector(d, f)
        assert f.shape == (nn, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = nVector(d, [2., 2.])
        with pytest.raises(AssertionError):
            f = nVector(d, [[2.], [1]])
        f = nVector(d, 'test')
        assert f.dtype in ('S4', '<U4')

    def test_tensor_field_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nTensor(d)
        assert type(f) == nTensor
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nTensor(d, 1.)
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        f = nTensor(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f[..., -1, 0], 2.)
        assert f.domain is d

        f = nTensor(d, f)
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f[..., 0, 0], 1.)
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = nTensor(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = nTensor(d, [[2., 1]])
        f = nTensor(d, 3)
        assert f.dtype == int

    def test_tensor3_field_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nTensor3(d)
        assert type(f) == nTensor3
        assert f.shape == (nn, nd, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nTensor3(d, 1.)
        assert f.shape == (nn, nd, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        r = np.random.rand(3, 3, 3)
        f = nTensor3(d, r)
        assert f.shape == (nn, nd, nd, nd)
        assert np.allclose(f[..., 1, 1, 1], r[1, 1, 1])
        assert f.domain is d

        f = nTensor3(d, f)
        assert f.shape == (nn, nd, nd, nd)
        assert np.allclose(f[..., 0, 0, 0], r[0, 0, 0])
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = nTensor3(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = nTensor3(d, [[2., 1]])
        f = nTensor3(d, 3)
        assert f.dtype == int

    def test_tensor4_field_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nTensor4(d)
        assert type(f) == nTensor4
        assert f.shape == (nn, nd, nd, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nTensor4(d, 1.)
        assert f.shape == (nn, nd, nd, nd, nd)
        assert np.allclose(f, 1.)
        assert f.domain is d

        r = np.random.rand(3, 3, 3, 3)
        f = nTensor4(d, r)
        assert f.shape == (nn, nd, nd, nd, nd)
        assert np.allclose(f[..., 1, 1, 1, 1], r[1, 1, 1, 1])
        assert f.domain is d

        f = nTensor4(d, f)
        assert f.shape == (nn, nd, nd, nd, nd)
        assert np.allclose(f[..., 0, 0, 0, 0], r[0, 0, 0, 0])
        assert f.domain is d

        with pytest.raises(AssertionError):
            f = nTensor4(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = nTensor4(d, [[2., 1]])
        f = nTensor4(d, 3)
        assert f.dtype == int

    def test_list_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nList(d, [1., 1., 0.])
        assert type(f) == nVector
        assert f.shape == (nn, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = nList(d, nVector(d, [1., 1., 0.]))
        assert type(f) == nVector
        assert f.shape == (nn, nd)
        assert np.allclose(f[..., -1], 0.)
        assert f.domain is d

        f = nList(d, 5)
        assert type(f) == nList
        assert f.shape == (nn, 5)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nList(d, [3, 2])
        assert type(f) == nList
        assert f.shape == (nn, 2)
        assert np.allclose(f[..., 0], 3)
        assert f.domain is d
        assert f.dtype == int

        with pytest.raises(AssertionError):
            f = nList(d, [[2., 2.], [1., 1.]])
        with pytest.raises(AssertionError):
            f = nList(d, [[2., 1]])
        with pytest.raises(AssertionError):
            f = nList(d, 1.)

    def test_matrix_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        f = nMatrix(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert type(f) == nTensor
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f[..., -1, 0], 2.)
        assert f.domain is d

        f = nMatrix(d, nTensor(d))
        assert type(f) == nTensor
        assert f.shape == (nn, nd, nd)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nMatrix(d, (5, 2))
        assert type(f) == nMatrix
        assert f.shape == (nn, 5, 2)
        assert np.allclose(f, 0.)
        assert f.domain is d

        f = nMatrix(d, [[3, 2], [2, 1]])
        assert type(f) == nMatrix
        assert f.shape == (nn, 2, 2)
        assert np.allclose(f[..., 0, 0], 3)
        assert f.domain is d
        assert f.dtype == int

        with pytest.raises(AssertionError):
            f = nMatrix(d, [2., 1])
        with pytest.raises(AssertionError):
            f = nMatrix(d, [[[2., 1]]])
        with pytest.raises(AssertionError):
            f = nMatrix(d, 1.)

    def test_nfield_creation(self):
        d = self.domain
        nn = d.nnode
        nd = d.ndime

        with pytest.raises(TypeError):
            f = nField(d)

        f = nField(d, 2.)
        assert type(f) == nScalar

        f = nField(d, [1, 2, 3])
        assert type(f) == nVector
        assert f.dtype == int

        f = nField(d, [1, 2, 3, 4.])
        assert type(f) == nList

        f = nField(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])
        assert type(f) == nTensor

        f = nField(d, [[1., 1., 0.], [1., 1., 0.]])
        assert type(f) == nMatrix

        r = np.random.rand(3, 3, 3)
        f = nField(d, r)
        assert type(f) == nTensor3

        r = np.random.rand(3, 3, 3, 3)
        f = nField(d, r)
        assert type(f) == nTensor4

        r = np.random.rand(3, 2, 3, 4)
        f = nField(d, r)
        assert type(f) == nField

        r = np.random.rand(nn, nd, nd)
        f = nField(d, r)
        assert type(f) == nTensor

    def test_tensordot_dot(self):
        d = self.domain
        t4 = nTensor4(d, 1.)
        t2 = nTensor(d, 3.)
        v = nVector(d, 4.)

        assert type(tensordot(t4, t2)) == nTensor
        assert type(tensordot(t2, t2)) == nScalar
        assert type(dot(t2, v)) == nVector
        assert type(dot(v, v)) == nScalar
        assert type(dot(t2, t2)) == nTensor

    def test_ufuncs(self):
        d = self.domain
        f = nField(d, [2, 3, 4.])  # scalar
        assert isinstance(np.sum(f), float)
        f = nField(d, [[1., 1., 0.], [1., 1., 0.], [2., 2., 1.]])  # tensor
        assert isinstance(np.sum(f, 2), nVector)
        assert isinstance(np.sum(f, f.saxes), nScalar)
        assert isinstance(np.sum(f, 0), np.ndarray)


if __name__ == '__main__':
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 10 16:20:15 2014

@author: mwojc
"""
import numpy as np


class Element(object):
    """
    Base finite element from which all elements should derive

    Subclasses have to provide the following attributes:
        ncoors          - local coordinates of nodes
        gcoors          - local coordinates of Gauss points
        gweights        - weights of Gauss points
        faces           - array of nodes defining faces or None
        btype           - type of face element or None

    and methods:
        shp(xloc)
        shpd(xloc)
        shpd2(xloc) [optional]
    """
    name = None
    shpd2 = None
    gshpd2 = None

    faces = None
    btype = None
    bncoors = None
    bgcoors = None
    bgshp = None
    bgshpd = None
    bgshpd2 = None
    nfaces = None

    def __init__(self):
        gc = self.gcoors
        self.nshp = np.array([self.shp(xloc) for xloc in self.ncoors])
        self.gshp = np.array([self.shp(xloc) for xloc in gc])
        self.gshpd = np.array([self.shpd(xloc) for xloc in gc])
        if self.shpd2 is not None:
            self.gshpd2 = np.array([self.shpd2(xloc) for xloc in gc])
        self.ngauss, self.nnode, self.ndime = self.gshpd.shape
        # boundary data...
        if self.faces is not None and self.btype is not None:
            if not isinstance(self.btype, Element):
                self.btype = self.btype()
            self.bncoors = self.ncoors[self.faces]
            self.bgcoors = np.einsum('eni, gn -> egi',
                                     self.bncoors, self.btype.gshp)
            self.bgshp = np.array([[self.shp(xloc) for xloc in bgc]
                                  for bgc in self.bgcoors])
            self.bgshpd = np.array([[self.shpd(xloc) for xloc in bgc]
                                   for bgc in self.bgcoors])
            if self.shpd2 is not None:
                self.bgshpd2 = np.array([[self.shpd2(xloc) for xloc in bgc]
                                        for bgc in self.bgcoors])
            self.nfaces = len(self.faces)


class Point10(Element):
    """
    1-node dummy point element
    """
    # Name used throughout fempy
    name = 'point'
    # Coordinates of nodes
    ncoors = np.array([[0.]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[0.]])
    gweights = np.array([1.])
    # Face ordering
    faces = None
    # Face element type
    btype = None

    # Shape functions...
    def shp(self, xloc):
        shp = np.zeros(1)
        shp[0] = 1.
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        dshp = np.array(()).reshape((1, 0))
        return dshp


class Line21(Element):
    """
    2-noded line with 1 Gauss point
    """
    # Name used throughout fempy
    name = 'line2'
    # Coordinates of nodes
    ncoors = np.array([[-1.], [1.]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[0.]])
    gweights = np.array([2.])
    # Face ordering
    faces = np.array([[0], [1]])
    # Face element type
    btype = Point10
    # Element segments (specific for 1D elements only)
    segments = [[0, 1]]

    # Shape functions...
    def shp(self, xloc):
        xi, = xloc
        shp = np.zeros(2)
        shp[0] = 0.5*(1.0-xi)
        shp[1] = 0.5*(1.0+xi)
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, = xloc
        dshp = np.zeros((2, 1))
        dshp[0, 0] = -0.5
        dshp[1, 0] = 0.5
        return dshp


class Line32(Element):
    """
    3-noded line with 2 Gauss points
    """
    # Name used throughout fempy
    name = 'line3'
    # Coordinates of nodes
    ncoors = np.array([[-1.], [1.], [0.]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[-0.577350269189626], [0.577350269189626]])
    gweights = np.array([1., 1.])
    # Face ordering
    faces = np.array([[0], [1]])
    # Face element type
    btype = Point10
    # Element segments (specific for 1D elements only)
    segments = [[0, 2], [2, 1]]

    # Shape functions...
    def shp(self, xloc):
        xi, = xloc
        shp = np.zeros(3)
        shp[0] = 0.5*xi*(xi-1.0)
        shp[1] = 0.5*xi*(xi+1.0)
        shp[2] = 1.0 - xi*xi
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, = xloc
        dshp = np.zeros((3, 1))
        dshp[0, 0] = xi-0.5
        dshp[1, 0] = xi+0.5
        dshp[2, 0] = -2.0*xi
        return dshp

    # ...and their second derivatives
    def shpd2(self, xloc):
        xi, = xloc
        dshp2 = np.zeros((3, 1, 1))
        dshp2[0, 0, 0] = 1.
        dshp2[1, 0, 0] = 1.
        dshp2[2, 0, 0] = -2.
        return dshp2


class Line32a(Element):
    """
    3-noded line with 2 Gauss points
    """
    # Name used throughout fempy
    name = 'line3a'
    # Coordinates of nodes
    ncoors = np.array([[-1.], [1.], [0.]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[-0.577350269189626], [0.577350269189626]])
    gweights = np.array([1., 1.])
    # Face ordering
    faces = np.array([[0], [1]])
    # Face element type
    btype = Point10
    # Element segments (specific for 1D elements only)
    segments = [[0, 2], [2, 1]]

    # Shape functions...
    def shp(self, xloc):
        xi, = xloc
        shp = np.zeros(3)
        shp[0] = 0.5*(1.0-xi)
        shp[1] = 0.5*(1.0+xi)
        shp[2] = 0.
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, = xloc
        dshp = np.zeros((3, 1))
        dshp[0, 0] = -0.5
        dshp[1, 0] = 0.5
        dshp[2, 0] = 0.
        return dshp


class Tria31(Element):
    """
    3-noded triangle with 1 Gauss point
    """
    # Name used throughout fempy
    name = 'tria3'
    # Coordinates of nodes
    ncoors = np.array([[0., 0.], [1., 0.], [0., 1.]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[0.333333333333333, 0.333333333333333]])
    gweights = np.array([0.5])
    # Face ordering
    faces = np.array([[0, 1], [1, 2], [2, 0]])
    # Face element type
    btype = Line21

    # Shape functions...
    def shp(self, xloc):
        xi, et = xloc
        shp = np.zeros(3)
        shp[0] = 1.-xi-et
        shp[1] = xi
        shp[2] = et
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, et = xloc
        dshp = np.zeros((3, 2))
        dshp[0, 0] = -1.
        dshp[1, 0] = 1.
        dshp[2, 0] = 0.
        dshp[0, 1] = -1.
        dshp[1, 1] = 0.
        dshp[2, 1] = 1.
        return dshp


class Tria63(Element):
    """
    6-noded triangle with 3 Gauss points
    """
    # Name used throughout fempy
    name = 'tria6'
    ncoors = np.array([[0., 0.], [1., 0.], [0., 1.],
                       [.5, 0.], [.5, .5], [0., .5]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[0.166666666666667, 0.166666666666667],
                       [0.666666666666667, 0.166666666666667],
                       [0.166666666666667, 0.666666666666667]])
    gweights = np.array([0.166666666666666, 0.166666666666666,
                         0.166666666666666])
    # Face ordering
    faces = np.array([[0, 1, 3], [1, 2, 4], [2, 0, 5]])
    # Face element type
    btype = Line32

    # Shape functions...
    def shp(self, xloc):
        xi, et = xloc
        shp = np.zeros(6)
        shp[0] = 2.*(0.5-xi-et)*(1.-xi-et)
        shp[1] = xi*(2.*xi-1.)
        shp[2] = et*(2.*et-1.)
        shp[3] = 4.*xi*(1.-xi-et)
        shp[4] = 4.*xi*et
        shp[5] = 4.*et*(1.-xi-et)
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, et = xloc
        dshp = np.zeros((6, 2))
        dshp[0, 0] = 4.*(xi+et)-3.
        dshp[1, 0] = 4.*xi-1.
        dshp[2, 0] = 0.
        dshp[3, 0] = 4.*(1.-2.*xi-et)
        dshp[4, 0] = 4.*et
        dshp[5, 0] = -4.*et
        dshp[0, 1] = 4.*(xi+et)-3.
        dshp[1, 1] = 0.
        dshp[2, 1] = 4.*et-1.
        dshp[3, 1] = -4.*xi
        dshp[4, 1] = 4.*xi
        dshp[5, 1] = 4.*(1.-xi-2.*et)
        return dshp

    # ...and their second derivatives
    def shpd2(self, xloc):
        xi, et = xloc
        dshp2 = np.zeros((6, 2, 2))
        dshp2[0, 0, 0] = 4.
        dshp2[1, 0, 0] = 4.
        dshp2[2, 0, 0] = 0.
        dshp2[3, 0, 0] = -8.
        dshp2[4, 0, 0] = 0.
        dshp2[5, 0, 0] = 0.
        dshp2[0, 1, 0] = 4.
        dshp2[1, 1, 0] = 0.
        dshp2[2, 1, 0] = 0.
        dshp2[3, 1, 0] = -4.
        dshp2[4, 1, 0] = 4.
        dshp2[5, 1, 0] = -4.
        dshp2[0, 0, 1] = 4.
        dshp2[1, 0, 1] = 0.
        dshp2[2, 0, 1] = 0.
        dshp2[3, 0, 1] = -4.
        dshp2[4, 0, 1] = 4.
        dshp2[5, 0, 1] = -4.
        dshp2[0, 1, 1] = 4.
        dshp2[1, 1, 1] = 0.
        dshp2[2, 1, 1] = 4.
        dshp2[3, 1, 1] = 0.
        dshp2[4, 1, 1] = 0.
        dshp2[5, 1, 1] = -8.
        return dshp2


class Tria63a(Element):
    """
    6-noded triangle with 3 Gauss points
    """
    # Name used throughout fempy
    name = 'tria6a'
    ncoors = np.array([[0., 0.], [1., 0.], [0., 1.],
                       [.5, 0.], [.5, .5], [0., .5]])
    # Coordinates and weights of Gauss points
    gcoors = np.array([[0.166666666666667, 0.166666666666667],
                       [0.666666666666667, 0.166666666666667],
                       [0.166666666666667, 0.666666666666667]])
    gweights = np.array([0.166666666666666, 0.166666666666666,
                         0.166666666666666])
    # Face ordering
    faces = np.array([[0, 1, 3], [1, 2, 4], [2, 0, 5]])
    # Face element type
    btype = Line32a

    # Shape functions...
    def shp(self, xloc):
        xi, et = xloc
        shp = np.zeros(6)
        shp[0] = 1.-xi-et
        shp[1] = xi
        shp[2] = et
        shp[3] = 0.
        shp[4] = 0.
        shp[5] = 0.
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, et = xloc
        dshp = np.zeros((6, 2))
        dshp[0, 0] = -1.
        dshp[1, 0] = 1.
        dshp[2, 0] = 0.
        dshp[3, 0] = 0.
        dshp[4, 0] = 0.
        dshp[5, 0] = 0.
        dshp[0, 1] = -1.
        dshp[1, 1] = 0.
        dshp[2, 1] = 1.
        dshp[3, 1] = 0.
        dshp[4, 1] = 0.
        dshp[5, 1] = 0.
        return dshp


class Tetr41(Element):
    """
    4-noded tetrahedra with 1 Gauss points
    """
    # Name used throughout fempy
    name = 'tetr4'
    # Coordinates of nodes
    ncoors = np.array([[0., 0., 0.], [1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    # Coordinates and weights of Gauss points
    gcoors = \
        np.array([[0.250000000000000, 0.250000000000000, 0.250000000000000]])
    gweights = np.array([0.166666666666667])
    # Face ordering
    faces = np.array([[0, 1, 3], [1, 2, 3], [2, 0, 3], [0, 2, 1]])
    # Face element type
    btype = Tria31

    # Shape functions...
    def shp(self, xloc):
        xi, et, ze = xloc
        shp = np.zeros(4)
        shp[0] = 1.-xi-et-ze
        shp[1] = xi
        shp[2] = et
        shp[3] = ze
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, et, ze = xloc
        dshp = np.zeros((4, 3))
        dshp[0, 0] = -1.
        dshp[1, 0] = 1.
        dshp[2, 0] = 0.
        dshp[3, 0] = 0.
        dshp[0, 1] = -1.
        dshp[1, 1] = 0.
        dshp[2, 1] = 1.
        dshp[3, 1] = 0.
        dshp[0, 2] = -1.
        dshp[1, 2] = 0.
        dshp[2, 2] = 0.
        dshp[3, 2] = 1.
        return dshp


class Tetr104(Element):
    """
    10-noded tetrahedra with 4 Gauss points
    """
    # Name used throughout fempy
    name = 'tetr10'
    # Coordinates of nodes
    ncoors = np.array([[0., 0., 0.], [1., 0., 0.], [0., 1., 0.], [0., 0., 1.],
                       [.5, 0., 0.], [.5, .5, 0.], [0., .5, 0.],
                       [0., 0., .5], [0., .5, .5], [.5, 0., .5]])
    # Coordinates and weights of Gauss points
    gcoors = \
        np.array([[0.138196601125011, 0.138196601125011, 0.138196601125011],
                  [0.585410196624968, 0.138196601125011, 0.138196601125011],
                  [0.138196601125011, 0.585410196624968, 0.138196601125011],
                  [0.138196601125011, 0.138196601125011, 0.585410196624968]])
    gweights = \
        np.array([0.041666666666667, 0.041666666666667, 0.041666666666667,
                  0.041666666666667])
    # Face ordering
    faces = np.array([[0, 1, 3, 4, 9, 7], [1, 2, 3, 5, 8, 9],
                      [2, 0, 3, 6, 7, 8], [0, 2, 1, 6, 5, 4]])
    # Face element type
    btype = Tria63

    # Shape functions...
    def shp(self, xloc):
        xi, et, ze = xloc
        a = 1.-xi-et-ze
        shp = np.zeros(10)
        shp[0] = (2.*a-1.)*a
        shp[1] = xi*(2.*xi-1.)
        shp[2] = et*(2.*et-1.)
        shp[3] = ze*(2.*ze-1.)
        shp[4] = 4.*xi*a
        shp[5] = 4.*xi*et
        shp[6] = 4.*et*a
        shp[7] = 4.*ze*a
        shp[8] = 4.*et*ze
        shp[9] = 4.*xi*ze
        return shp

    # ...and their derivatives
    def shpd(self, xloc):
        xi, et, ze = xloc
        dshp = np.zeros((10, 3))
        dshp[0, 0] = 1.-4.*(1.-xi-et-ze)
        dshp[1, 0] = 4.*xi-1.
        dshp[2, 0] = 0.
        dshp[3, 0] = 0.
        dshp[4, 0] = 4.*(1.-2.*xi-et-ze)
        dshp[5, 0] = 4.*et
        dshp[6, 0] = -4.*et
        dshp[7, 0] = -4.*ze
        dshp[8, 0] = 0.
        dshp[9, 0] = 4.*ze
        dshp[0, 1] = 1.-4.*(1.-xi-et-ze)
        dshp[1, 1] = 0.
        dshp[2, 1] = 4.*et-1.
        dshp[3, 1] = 0.
        dshp[4, 1] = -4.*xi
        dshp[5, 1] = 4.*xi
        dshp[6, 1] = 4.*(1.-xi-2.*et-ze)
        dshp[7, 1] = -4.*ze
        dshp[8, 1] = 4.*ze
        dshp[9, 1] = 0.
        dshp[0, 2] = 1.-4.*(1.-xi-et-ze)
        dshp[1, 2] = 0.
        dshp[2, 2] = 0.
        dshp[3, 2] = 4.*ze-1.
        dshp[4, 2] = -4.*xi
        dshp[5, 2] = 0.
        dshp[6, 2] = -4.*et
        dshp[7, 2] = 4.*(1.-xi-et-2.*ze)
        dshp[8, 2] = 4.*et
        dshp[9, 2] = 4.*xi
        return dshp

    # ...and their second derivatives
    def shpd2(self, xloc):
        xi, et, ze = xloc
        dshp2 = np.zeros((10, 3, 3))
        dshp2[0, 0, 0] = 4.
        dshp2[1, 0, 0] = 4.
        dshp2[2, 0, 0] = 0.
        dshp2[3, 0, 0] = 0.
        dshp2[4, 0, 0] = -8.
        dshp2[5, 0, 0] = 0.
        dshp2[6, 0, 0] = 0.
        dshp2[7, 0, 0] = 0.
        dshp2[8, 0, 0] = 0.
        dshp2[9, 0, 0] = 0.
        dshp2[0, 1, 0] = 4.
        dshp2[1, 1, 0] = 0.
        dshp2[2, 1, 0] = 0.
        dshp2[3, 1, 0] = 0.
        dshp2[4, 1, 0] = -4.
        dshp2[5, 1, 0] = 4.
        dshp2[6, 1, 0] = -4.
        dshp2[7, 1, 0] = 0.
        dshp2[8, 1, 0] = 0.
        dshp2[9, 1, 0] = 0.
        dshp2[0, 2, 0] = 4.
        dshp2[1, 2, 0] = 0.
        dshp2[2, 2, 0] = 0.
        dshp2[3, 2, 0] = 0.
        dshp2[4, 2, 0] = -4.
        dshp2[5, 2, 0] = 0.
        dshp2[6, 2, 0] = 0.
        dshp2[7, 2, 0] = -4.
        dshp2[8, 2, 0] = 0.
        dshp2[9, 2, 0] = 4.
        dshp2[0, 0, 1] = 4.
        dshp2[1, 0, 1] = 0.
        dshp2[2, 0, 1] = 0.
        dshp2[3, 0, 1] = 0.
        dshp2[4, 0, 1] = -4.
        dshp2[5, 0, 1] = 4.
        dshp2[6, 0, 1] = -4.
        dshp2[7, 0, 1] = 0.
        dshp2[8, 0, 1] = 0.
        dshp2[9, 0, 1] = 0.
        dshp2[0, 1, 1] = 4.
        dshp2[1, 1, 1] = 0.
        dshp2[2, 1, 1] = 4.
        dshp2[3, 1, 1] = 0.
        dshp2[4, 1, 1] = 0.
        dshp2[5, 1, 1] = 0.
        dshp2[6, 1, 1] = -8.
        dshp2[7, 1, 1] = 0.
        dshp2[8, 1, 1] = 0.
        dshp2[9, 1, 1] = 0.
        dshp2[0, 2, 1] = 4.
        dshp2[1, 2, 1] = 0.
        dshp2[2, 2, 1] = 0.
        dshp2[3, 2, 1] = 0.
        dshp2[4, 2, 1] = 0.
        dshp2[5, 2, 1] = 0.
        dshp2[6, 2, 1] = -4.
        dshp2[7, 2, 1] = -4.
        dshp2[8, 2, 1] = 4.
        dshp2[9, 2, 1] = 0.
        dshp2[0, 0, 2] = 4.
        dshp2[1, 0, 2] = 0.
        dshp2[2, 0, 2] = 0.
        dshp2[3, 0, 2] = 0.
        dshp2[4, 0, 2] = -4.
        dshp2[5, 0, 2] = 0.
        dshp2[6, 0, 2] = 0.
        dshp2[7, 0, 2] = -4.
        dshp2[8, 0, 2] = 0.
        dshp2[9, 0, 2] = 4.
        dshp2[0, 1, 2] = 4.
        dshp2[1, 1, 2] = 0.
        dshp2[2, 1, 2] = 0.
        dshp2[3, 1, 2] = 0.
        dshp2[4, 1, 2] = 0.
        dshp2[5, 1, 2] = 0.
        dshp2[6, 1, 2] = -4.
        dshp2[7, 1, 2] = -4.
        dshp2[8, 1, 2] = 4.
        dshp2[9, 1, 2] = 0.
        dshp2[0, 2, 2] = 4.
        dshp2[1, 2, 2] = 0.
        dshp2[2, 2, 2] = 0.
        dshp2[3, 2, 2] = 4.
        dshp2[4, 2, 2] = 0.
        dshp2[5, 2, 2] = 0.
        dshp2[6, 2, 2] = 0.
        dshp2[7, 2, 2] = -8.
        dshp2[8, 2, 2] = 0.
        dshp2[9, 2, 2] = 0.


# Convenience definitions
etypes = {'point': Point10,
          'line2': Line21,
          'line3': Line32,
          'tria3': Tria31,
          'tria6': Tria63,
          'tetr4': Tetr41,
          'tetr10': Tetr104}


edims = {(0, 1):  etypes['point'],
         (1, 2):  etypes['line2'],
         (1, 3):  etypes['line3'],
         (2, 3):  etypes['tria3'],
         (2, 6):  etypes['tria6'],
         (3, 4):  etypes['tetr4'],
         (3, 10): etypes['tetr10']
         }


def guess_etype(coors, conec, etype=None):
    enode = conec.shape[1]
    if isinstance(etype, Element):
        assert enode == etype.nnode, \
            "Conec requires %s-node element but %s is defined on %s" \
            " nodes." % (enode, etype.__name__, etype.nnode)
        return etype
    try:
        if isinstance(etype, type) and issubclass(etype, Element):
            etype = etype()
            assert enode == etype.nnode, \
                "Conec requires %s-node element but %s is defined on %s" \
                " nodes." % (enode, etype.__name__, etype.nnode)
            return etype
    except TypeError:
        pass
    if isinstance(etype, str) and etype in etypes:
        return etypes[etype]()
    if isinstance(etype, int):  # etype is treated as edime
        try:
            return edims[(etype, enode)]()
        except KeyError:
            raise ValueError("%iD element with %i nodes not found!"
                             % (etype, enode))
    if etype is None:
        if enode == 1:  # point
            etype = Point10
        elif enode == 2:  # line2
            etype = Line21
        elif enode == 3:
            if len(np.intersect1d(conec[:, :2], conec[:, 2])) == 0:  # line3
                # TODO: doesn't work for single element...
                etype = Line32
            else:
                etype = Tria31
        elif enode == 4:
            etype = Tetr41  # tetr4
        elif enode == 6:  # tria6
            etype = Tria63
        elif enode == 10:  # tetr10
            etype = Tetr104
        else:
            raise ValueError("Element type is not recognized!")
        return etype()
    raise ValueError("Unknown element type: %s" % etype)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')




# -*- coding: utf-8 -*-
"""
Created on Mon Jun  5 15:03:58 2017

@author: mwojc
"""
import numpy as np


class cached_property(object):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/
        commit/fa7733e075da0d790d809aa3d2f53071897e6f76

    This class can be used as a decorator replacing the built-in python
    'property'
    """
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def invalidate_caches(obj, names=None):
    """
    Searches for cached_properties in 'obj' and resets them.
    """
    # get all names in model class and object
    c = obj.__class__
    d = obj.__dict__
    cdir = dir(c) if names is None else names
    dkeys = d.keys()
    for n in cdir:
        # invalidate cached_properties if they have been used...
        if isinstance(getattr(c, n), cached_property) and n in dkeys:
            # property has been cached so we invalidate it
            del d[n]


def renumber(coors, conec):
    """
    Checks which coors are used by conec, and then renumber conec.

    Returns a tuple: (new_coors, new_conec, nlist), where nlist contains
    node numbers in original coors used by new_conec. The following is true
    for the returned values:
        new_coors = coors[nlist]
        conec = nlist[new_conec]
    """
    maxn = conec.max()
    if maxn > len(coors) - 1:
        raise ValueError("No such node: %i" % maxn)
    nlist = np.unique(conec)
    if len(nlist) < len(coors):
        cnlist = np.zeros(len(coors), dtype=int) - 1
        cnlist[nlist] = np.arange(len(nlist), dtype=int)
        coors = coors[nlist]
        conec = cnlist[conec]
    return coors, conec, nlist


def uneven_sum(arr, mask, keepshape=False):
    """
    Sums 'arr' along 0-th axis with slices defined by boolean
    'mask'. If 'keepshape' is True the return array has the
    same dimension as 'arr' (with repeated entries)

    Note: similar functionality is achieved with

    >>> np.add.reduceat(arr, np.nonzero(mask)[0])
    """
    idx = np.cumsum(mask) - 1
    res_shape = (mask.sum(),) + arr.shape[1:]
    res = np.zeros(res_shape, dtype=arr.dtype)
    np.add.at(res, idx, arr)
    if keepshape:
        return res[idx]
    return res


def unique_rows_lex(a, sort=False, strict=True):
    """
    Sorts lexically 2D integer array and searches for duplicate rows

    Arguments:
        a       - 2D integer array
        sort    - if True, rows in 'a' are sorted first (default is False)
        strict  - if True (default) only the rows which appear once in 'a' are
                  treated as uniqe, if False, also the first occurence of
                  duplicated rows is added to the resulting mask

    Returns tuple (idx, mask) is returned where 'idx' are indices of lexical
    sort of 'a' and 'mask' is a boolean array pointing out the unique rows
    numbers in idx.
    """
    if len(a) == 0:
        return np.array([], dtype=np.int), np.array([], dtype=np.bool)
    a = np.asarray(a)
    if sort:
        a = np.sort(a)      # sort rows
    idx = np.lexsort(a.T)   # find indices which sort along last axis
    sa = a[idx]             # get lexically sorted version of a
    # mask is False if previous entry in sa is the same
    mask = np.append([True, ], np.any(np.diff(sa, axis=0), axis=1))
    if strict:
        # if mask entry is False make also previous entry False
        mask[np.nonzero(~mask)[0] - 1] = False
        # now only entries which appear once are marked True in mask
    return idx, mask


def unique_rows(a, sort=False, strict=True):
    """
    Compute unique rows in 2D integer array

    Arguments like in unique_rows_lex.

    Returns boolean mask which points the unique rows in 'a'.
    """
    idx, mask = unique_rows_lex(a, sort=sort, strict=strict)
    # recover mask for original array from its lexical sort
    idx = idx[mask]
    mask[:] = False
    mask[idx] = True
    return mask


def duplicate_rows(a, sort=False):
    """
    Computes tags for the duplicate rows in 2D array

    Arguments like in unique_rows_lex.

    Returns 'dups' - an integer array with the same integer tag for duplicate
    rows in 'a'. Tags are numbered from 1 to the number of strictly unique
    (possibly sorted first). The order of tags is ruled by lexical sort
    of 'a.T'
    """
    idx, mask = unique_rows_lex(a, sort=sort, strict=False)
    # mark duplicates with subsequent integers
    dups = np.cumsum(mask)
    d = np.zeros_like(dups)
    d[idx] = dups
    return d


def in2d(a, b, sort=False):
    """
    Test whether 'b' rows are present in 'a'.

    Returns boolean array of the same length as 'a' that is True where the
    row of 'a' is in 'b'. If sort=True the rows in both arrays are sorted
    first.
    """
    n = len(a)
    dp = duplicate_rows(np.concatenate([a, b]), sort=sort)
    return np.in1d(dp[:n], dp[n:])


find_faces = in2d


def agree_rows(a, b):
    """
    Agree indices of same rows in a and b (which are 2D arrays).

    For a given 'a' and 'b' computes index arrays 'aidx', 'bidx', such that
    a[aidx] == b[bidx]. Input arrays must be of the same lengts and all 'a'
    rows must be present in 'b' otherwise ValueError is raised

    Returns tuple (aidx, bidx)
    """
    aidx = np.lexsort(a.T)
    bidx = np.lexsort(b.T)
    if np.allclose(a[aidx], b[bidx]):
        return aidx, bidx
    else:
        raise ValueError("Rows indices cannot be agreed")

# def find_faces(fconec1, fconec2, edges=None, sort=False):
#    """
#    Finds positions of faces2 in faces1 as boolean array
#    If sort is given then orientation is also unified
#    """
#    n = len(fconec1)
#    join = np.concatenate
#    if edges is None or sort is True:
#        dp = duplicate_rows(join([fconec1, fconec2]), sort=sort)
#        return np.in1d(dp[:n], dp[n:])
#    e1 = fconec1[:, edges].reshape(-1, edges.shape[1])
#    e2 = fconec2[:, edges].reshape(-1, edges.shape[1])
#    dp = duplicate_rows(join([e1, e2]))
#    return np.in1d(dp[:n], dp[n:]).reshape(n, -1).all(1)


def normal_sign(domain):
    """
    +1 out, -1 in
    """
    d = domain
    if d.etype.btype is None or d.etype.ndime < 2:
        return
    from index import EIndex
    fconec = d.eidx.conec_g[:, d.etype.faces[0]]  # global face conec
    fidx = EIndex(fconec, nidx=d.eidx.nidx.g)
    faces = d.btype(fidx, d.etype.btype)
    n = faces.normal[:, 0]           # normal, only first Gauss point
    p = faces.ecoors[:, 0]           # any (first) face point
    v = d.ecoors_mean                # element center point
    s = np.sign(np.einsum('ij,ij->i', n, v-p)).astype(int)  # +1 out, -1 in
    return s


def create_faces(domain):
    d = domain
    # Do nothing if there are no boundary elements
    if d.etype.btype is None:
        return
    # Create all faces conec
    fconec0 = d.eidx.conec_g[:, d.etype.faces]         # global fconec
    fconec = fconec0.reshape(-1, fconec0.shape[-1])    # reshape to 2D
    fconec = fconec[unique_rows(fconec, strict=False)]
    # Create faces index
    from fempy.domain.index import EIndex
    fidx = EIndex(fconec, nidx=d.eidx.nidx)  # should fconec be unique?
    # Add the groups to the index
    for k, v in d.eidx.items():
        vcon = fconec0[v].reshape(-1, fconec0.shape[-1])  #
        vbool = find_faces(fconec, vcon, sort=True)
        # TODO: below, nvbool recovers ngroups overwritten by setting fidx[k]
        # they are overwritten because we use the same global node index for
        # domain and faces and when setting element groups this global node
        # index is also updated (in a hard way). We need a kind of dynamic
        # __getitem__ to always compute the result when eidx.nidx[k]
        # is called with 'k' present in eidx only
        if k in d.eidx.nidx:
            nvbool = d.eidx.nidx.g[k]  # [fidx.nidx.lst]
        fidx[k] = vbool
        if k in d.eidx.nidx:
            fidx.nidx.g[k] = nvbool
        # d.eidx[k] = v  # TODO: recovers ngroups overwritten in faces
    # Now handle boundary
    bbool = unique_rows(fconec, sort=True)
    if 'BOUNDARY' not in fidx.g:  # add global boundary...
        fidx['BOUNDARY'] = bbool
    # add locally to faces domain
    fidx.set_local_mask('boundary', bbool)
    fidx.set_local_mask('CUT', fidx['boundary^BOUNDARY'])
    # add locally to domain nidx
    d.eidx.nidx.set_local_mask('boundary', fidx.conec[bbool])
    d.eidx.nidx.set_local_mask('CUT', fidx.conec[fidx['CUT']])
    # Create faces
    faces = d.btype(fidx, d.etype.btype)
    return faces


# Tests
if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 16 13:49:09 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.fields import Field


class nField(Field):
    def __new__(cls, domain, data):
        if isinstance(data, nField):
            if domain is data.domain:
                return data
        nn = domain.nnode
        nd = domain.ndime
        data = np.asarray(data)
        prefix = 1
        if not data.shape[:prefix] == (nn,):
            default = data
            data = np.empty((nn,) + data.shape, dtype=data.dtype)
            data[:] = default
        # Field factory
        sshape = data.shape[prefix:]
        if sshape == ():
            cls = nScalar
        elif len(sshape) == 1 and sshape == (nd,):
            cls = nVector
        elif len(sshape) == 1 and sshape != (nd,):
            cls = nList
        elif len(sshape) == 2 and sshape == (nd,)*2:
            cls = nTensor
        elif len(sshape) == 2 and sshape != (nd,)*2:
            cls = nMatrix
        elif sshape == (nd,)*3:
            cls = nTensor3
        elif sshape == (nd,)*4:
            cls = nTensor4
        else:
            pass
        f = Field.__new__(cls, data, prefix=prefix)
        f.domain = domain
        return f

    def __array_finalize__(self, field):
        if field is None:
            return
        Field.__array_finalize__(self, field)
        self.domain = getattr(field, 'domain', None)

    @property
    def ndofs(self):
        return np.prod(self.sshape, dtype=int)

    @property
    def edofs(self):
        return self.domain.etype.nnode * self.ndofs

    @property
    def nsize(self):
        return self.domain.eidx.nidx.g.na * self.ndofs

    @property
    def _nipos(self):
        idx = self.domain.nlist
        return np.arange(self.nsize).reshape(-1, self.ndofs)[idx]

    @property
    def nipos(self):
        return self.factory(self._nipos)

    @property
    def eipos(self):
        return self._nipos[self.domain.conec].reshape(-1, self.edofs)

    @property
    def eij(self):
        as_strided = np.lib.stride_tricks.as_strided
        eipos = self.eipos
        m, n = eipos.shape
        sz = eipos.itemsize
        ei = as_strided(eipos, shape=(m, n, n), strides=(n*sz, sz, 0))
        ej = as_strided(eipos, shape=(m, n, n), strides=(n*sz, 0, sz))
        return ei, ej

    @property
    def groups(self):
        return self.domain.ngroups

    def __add__(self, other):
        if isinstance(other, nField):
            sd = self.domain
            od = other.domain
            if sd.eidx.nidx.g is od.eidx.nidx.g:  # the same space
                nlist = np.intersect1d(sd.eidx.nidx.lst,
                                       od.eidx.nidx.lst, assume_unique=True)
                sidx = sd.eidx.nidx.pos[nlist]
                oidx = od.eidx.nidx.pos[nlist]
                res = self.copy()
                res[sidx] += other[oidx]
                return res
            raise ValueError("Fields are defined of different point spaces")
        return Field.__add__(self, other)

    def __iadd__(self, other):
        if isinstance(other, nField):
            sd = self.domain
            od = other.domain
            if sd.eidx.nidx.g is od.eidx.nidx.g:  # the same space
                nlist = np.intersect1d(sd.eidx.nidx.lst,
                                       od.eidx.nidx.lst, assume_unique=True)
                sidx = sd.eidx.nidx.pos[nlist]
                oidx = od.eidx.nidx.pos[nlist]
                res = self
                res[sidx] += other[oidx]
                return res
            raise ValueError("Fields are defined of different point spaces")
        return Field.__add__(self, other)

    def factory(self, data=None, superdomain=False):
        if data is None:
            data = np.zeros(self.sshape)
        if superdomain:
            return nField(self.domain.superdomain, data)
        return nField(self.domain, data)


class nScalar(nField):
    def __new__(cls, domain, data=None):
        nn = domain.nnode
        if data is None:
            data = np.zeros((nn,))
        data = np.asarray(data)
        assert data.shape == (nn,) or data.shape == ()
        return nField.__new__(cls, domain, data)


class nList(nField):
    def __new__(cls, domain, data):
        nn = domain.nnode
        if isinstance(data, int):
            data = np.zeros((nn, data))
        else:
            data = np.asarray(data)
            assert len(data.shape) in (1, 2)
            if len(data.shape) == 2:
                assert data.shape[0] == nn
        return nField.__new__(cls, domain, data)


class nVector(nList):
    def __new__(cls, domain, data=None):
        nn = domain.nnode
        nd = domain.ndime
        if data is None:
            data = np.zeros((nn, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, nd)
        assert data.shape == (nn, nd) or data.shape == (nd,)
        return nField.__new__(cls, domain, data)


class nMatrix(nField):
    def __new__(cls, domain, data):
        nn = domain.nnode
        if isinstance(data, int):
            data = np.zeros((nn, data, data))
        else:
            data = np.asarray(data)
            assert len(data.shape) in (1, 2, 3)
            if len(data.shape) == 1:
                assert data.dtype == int and data.shape[0] == 2
                data = np.zeros(tuple(data))
            if len(data.shape) == 3:
                assert data.shape[0] == nn
        return nField.__new__(cls, domain, data)


class nTensor(nMatrix):
    def __new__(cls, domain, data=None):
        nn = domain.nnode
        nd = domain.ndime
        if data is None:
            data = np.zeros((nn, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd))
        assert data.shape == (nn, nd, nd) or data.shape == (nd,)*2
        return nField.__new__(cls, domain, data)


class nTensor3(nField):
    def __new__(cls, domain, data=None):
        nn = domain.nnode
        nd = domain.ndime
        if data is None:
            data = np.zeros((nn, nd, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd, nd))
        assert data.shape == (nn, nd, nd, nd) or data.shape == (nd,)*3
        return nField.__new__(cls, domain, data)


class nTensor4(nField):
    def __new__(cls, domain, data=None):
        nn = domain.nnode
        nd = domain.ndime
        if data is None:
            data = np.zeros((nn, nd, nd, nd, nd))
        data = np.asarray(data)
        if data.shape == ():
            data = np.tile(data, (nd, nd, nd, nd))
        assert data.shape == (nn, nd, nd, nd, nd) or data.shape == (nd,)*4
        return nField.__new__(cls, domain, data)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 12:54:09 2016

@author: mwojc
"""

import pytest
import numpy as np
import scipy.sparse as sparse
import fempy.solvers.precon as precon

pn = [p for p in dir(precon) if 'Preconditioner' in p]
pc = [getattr(precon, p) for p in pn]
# po = [p() for p in pc]


class Test_Preconditioners:
    def get_dense_sparse_matrix(self, n, symmetric=False):
        Ad = np.random.rand(n, n)
        if symmetric:
            Ad = (Ad + Ad.T)/2.
        As = sparse.csr_matrix(Ad)
        return Ad, As

    def get_rhs(self, n):
        return np.random.rand(n)

    def setup(self):
        n = 10
        Ad, As = self.get_dense_sparse_matrix(n)
        b = self.get_rhs(n)
        x = np.linalg.solve(Ad, b)
        self.Ad = Ad
        self.As = As
        self.b = b
        self.x = x

    def test_superlu_preconditioner(self):
        p = precon.lu()
        x = p(self.Ad)*self.b
        assert np.allclose(x, self.x)
        x = p(self.As)*self.b
        assert np.allclose(x, self.x)

    # @pytest.mark.skipif('UmfpackPreconditioner' not in pn)
    def test_umfpack_preconditioner(self):
        pytest.importorskip('scikits.umfpack')
        p = precon.umfpack()
        x = p(self.Ad)*self.b
        assert np.allclose(x, self.x)
        x = p(self.As)*self.b
        assert np.allclose(x, self.x)

    # @pytest.mark.skipif('AmgPrecondtioner' not in pn, reason='no pyamg')
    def test_amg_preconditioner(self):
        pytest.importorskip('pyamg')
        p = precon.amg_sas()
        x1 = p(self.Ad)*self.b
        x2 = p(self.As)*self.b
        assert np.allclose(x1, x2)

    # @pytest.mark.skipif('AmgPrecondtioner' not in pn, reason='no pyamg')
    def test_amgcl_preconditioner(self):
        pytest.importorskip('pyamgcl')
        p = precon.amgcl_cg()
        x1 = p(self.Ad)*self.b
        x2 = p(self.As)*self.b
        assert np.allclose(x1, x2)

    def test_superilu_preconditioner(self):
        p = precon.ilu()
        x1 = p(self.Ad)*self.b
        x2 = p(self.As)*self.b
        assert np.allclose(x1, x2)

    def test_diag_preconditioner(self):
        p = precon.diag()
        x1 = p(self.Ad)*self.b
        x2 = p(self.As)*self.b
        assert np.allclose(x1, x2)


if __name__ == '__main__':
    # import pytest
    pytest.main([str(__file__), '-v'])  # Run tests from current file



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 21:32:10 2014

@author: mwojc
"""
import os
import sys
import pytest

filedir = os.path.dirname(__file__)
if os.path.isdir(os.path.join(filedir, 'tests')):
    testdir = os.path.realpath(filedir + '/tests') + '/'
    testdata = os.path.realpath(filedir + '/tests/data') + '/'
else:
    testdir = os.path.realpath(filedir + '/../tests') + '/'
    testdata = os.path.realpath(filedir + '/../tests/data') + '/'
sys.path.append(testdir)
# imports from testdir
import common
from common import *


def gettestfile(module):
    module = module.replace(filedir, '')
    p, s = os.path.split(module)
    s = s.replace('.pyc', '.py')
    testfile = p + '/test_' + s
    testfile = testdir + testfile
    open(testfile)  # for rising possible IOError
    return testfile


def testmodule(module, opts=''):
    try:
        testfile = gettestfile(module)
        d, f = os.path.split(testfile)
        curdir = os.getcwd()
        os.chdir(d)
        opts = opts if isinstance(opts, list) else opts.split()
        pytest.main([str(f)] + opts)
        os.chdir(curdir)
    except IOError:
        pass  # Skip testing silently if no testing file


if __name__ == '__main__':
    pytest.main([testdir, '-vrs'])  # Run all tests from testdir



# -*- coding: utf-8 -*-
"""
Created on Wed May 18 12:07:58 2016

@author: mwojc
"""
from fempy.models.model import Model
from fempy.fields.nfields import nScalar
from fempy.fields.gfields import gScalar, gTensor
from fempy.fields.operators import dot, grad, operator, onboundary, nabla
from fempy.fields.operators import qform, lform, integrate, ongauss, lformu
# import numpy as np

#from fempy.domain.tools import cached_property
#property = cached_property


class Heat(Model):
    def init(self, **kwargs):
        d = self.domain
        b = self.domain.boundary
        # on domain nodes
        self.T = nScalar(d, 273.15)     # temperature [K]
        self.bcs = nScalar(d, False)    # Dirichlet boundary conditions mask
        # on domain integration points
        self.kij = gTensor(d)           # heat conductivity        [W/m*K]
        self.cp = gScalar(d, 1.)        # heat capacity (isobaric) [J/kg*K]
        self.rho = gScalar(d, 1.)       # material density         [kg/m**3]
        self.h = gScalar(d)             # heat source              [W/kg]
        # on boundary
        self.q = gScalar(b)             # RHS explicit flux        [W/m**2]
        self.Text = gScalar(b, 273.15)  # external temperature on boundary [K]
        self.alpha = gScalar(b)         # heat transfer coefficient [W/m**2*K]
        self.beta = 5.67036713131e-8    # W/(m**2 * K**4) - Stefan-Boltzmann
        self.epsilon = gScalar(b)       # surface emissivity [-]

        # temperature is our unknown, the rest we define as properties
        self.lhs = self.T

    @property
    def gamma(self):
        Tb = ongauss(onboundary(self.T))
        Te = self.Text
        b = self.beta
        e = self.epsilon
        return (Tb**3 + Te*(Tb**2) + (Te**2)*Tb + Te**3) * b * e

    @property
    def rhs(self):
        T = self.T
        Tb = onboundary(T)
        # Assemble sources
        source = dot(self.rho, self.h)
        qs = lform(source, operator(T))
        # Assemble convective heat transfer on boundary
        transfer = dot(self.alpha, self.Text)  # ongauss(Tb) - self.Text)
        qt = lform(transfer, operator(Tb))
        # Assemble radiation at boundary
        radiation = dot(self.gamma, self.Text)
        qr = lform(radiation, operator(Tb))
        return lform(self.q) + qs + qt + qr

    @property
    def lop(self):
        L1 = -integrate(self.Q)
        L2 = lform(dot(self.alpha, ongauss(onboundary(self.T))))
        L3 = lform(dot(self.gamma, ongauss(onboundary(self.T))))
        return L1 + L2 + L3

    @property
    def jac(self):
        K1 = qform(self.kij, nabla(self.T))  # the same trial and test
        K2 = qform(self.alpha, operator(onboundary(self.T)))
        K3 = qform(self.gamma, operator(onboundary(self.T)))
        return K1 + K2 + K3

    @property
    def jact(self):
        return lformu(self.rho * self.cp)

    @property
    def gradT(self):
        return grad(self.T)

    @property
    def Q(self):
        return -dot(self.kij, self.gradT)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')
#    from fempy.geometry import Rectangle
#    d = Rectangle(0.3, 0.1, elsize_factor=0.5).gmsh.domain(order=1)
#
#    m = Heat(d)
#
#    # m.k[:] = 1. #[[1., 0], [0, 1.]]
#    m.kij[:] = [[1., 0], [0, 1.]]
#
#    m.alpha['left + bottom'] = 8.
#    m.epsilon['left + bottom'] = 0.85
#    m.Text['left + bottom'] = 293.15
#
#    m.alpha['right + top'] = 25.
#    m.epsilon['right + top'] = 0.85
#    m.Text['right + top'] = 1273.15
#
##    from fempy.solvers import ivp
##    m.solver = ivp.ODESolver
##    m.time0 = 0.
##    m.time1 = 1
##    m.T[:] = 293.15
#    m.solve(verbose=True)
#    m.preview()



# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 11:51:44 2016

@author: mwojc
"""


class Solver(object):
    default_options = None

    def __init__(self, **options):
        self.options = options

    def getopts(self, options=None):
        if self.default_options is not None:
            opts = self.default_options.copy()
        else:
            opts = {}
        if hasattr(self, 'options'):
            opts.update(self.options)
        if options is not None:
            opts.update(options)
        return opts

    def solve(self, *args, **options):
        raise NotImplementedError


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



#import tools
#from domain import Domain
#from domain import Volume
#from domain import Surface
#from domain import Surface3D
#from domain import Curve
#from domain import Curve2D
#from domain import Curve3D
#from domain import Points

# from index import GlobalNIndex
# from index import GlobalEIndex
# from index import NIndex
# from index import EIndex



# -*- coding: utf-8 -*-
"""
Created on Fri Jul  4 11:39:16 2014

@author: mwojc
"""
import numpy as np


class Field(np.ndarray):
    def __new__(cls, data, prefix=1, groups=None):
        data = np.asarray(data)
        shape = data.shape
        assert len(shape) >= prefix
        field = data.view(cls)
        field.prefix = prefix
        field.pshape = data.shape[:prefix]
        field.sshape = data.shape[prefix:]
        if not isinstance(groups, dict):
            groups = {}
        if not hasattr(cls, 'groups'):
            field.groups = groups
        return field

    def __array_finalize__(self, field):
        if field is None:
            return
        self.prefix = getattr(field, 'prefix', 1)
        self.pshape = getattr(field, 'pshape', ())
        self.sshape = getattr(field, 'sshape', ())
        if not hasattr(self.__class__, 'groups'):
            self.groups = getattr(field, 'groups', {})

    def __array_wrap__(self, out_arr, context=None):
        arr = out_arr.view(np.ndarray)
        if len(arr.shape) >= self.prefix:
            if arr.shape[:self.prefix] == self.pshape:
                arr = self.factory(arr)
        if arr.ndim == 0:
            arr = arr[()]
        return arr

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Field, self).__reduce__()
        # Create our own tuple to pass to __setstate__  (with custom __dict__)
        new_state = pickled_state[2] + (self.__dict__,)
        # Return a tuple that replaces the parent's __setstate__ tuple
        # with our own
        return (pickled_state[0], pickled_state[1], new_state)

    def __setstate__(self, state):
        self.__dict__.update(state[-1])  # Recover __dict__
        # Call the parent's __setstate__ with the other tuple elements.
        super(Field, self).__setstate__(state[0:-1])

    def __getslice__(self, i, j):
        return np.ndarray.__getslice__(self, i, j).view(np.ndarray)

    def __getitem__(self, key):
        key = self._get_key(key)
        return np.ndarray.__getitem__(self, key).view(np.ndarray)
        # TODO: get subdomain here?

    def __setitem__(self, key, value):
        key = self._get_key(key)
        return np.ndarray.__setitem__(self, key, value)

    def _get_key(self, key):
        if isinstance(key, str):
            key = (key,)
        if isinstance(key, tuple) and isinstance(key[0], str):
            key = (self.groups[key[0]], ) + key[1:]
        return key

    @property
    def sufix(self):
        return len(self.shape) - self.prefix

    @property
    def ssize(self):
        return np.prod(self.sshape, dtype=int)

    @property
    def paxes(self):
        return tuple(range(self.prefix))

    @property
    def saxes(self):
        return tuple(range(self.prefix, len(self.shape)))

    def factory(self, data=None):
        if data is None:
            data = np.zeros(self.sshape)
        return self.__class__(data, prefix=self.prefix, groups=self.groups)


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 19:09:58 2016

@author: mwojc
"""
from __future__ import print_function
import warnings
import scipy.sparse
from scipy.integrate import solve_ivp
from scipy.optimize.nonlin import KrylovJacobian
import numpy as np
from fempy.solvers import linear
from fempy.solvers import _bdf

hasattr = linear.hasattr


def _identity_like(A, n=None):
    if A is None:
        return
    if n is None:
        n = A.shape[0]
    if scipy.sparse.issparse(A):
        I = scipy.sparse.eye(n, dtype=A.dtype, format=A.format)
    else:  # is dense
        I = np.eye(n, dtype=A.dtype)
    return I


def _set_solution_at_time(m, t=None):
    if t is None:
        t = m._sol.t[-1]
    tmin = m._sol.t[0]
    tmax = m._sol.t[-1]
    if t < tmin or t > tmax:
        warnings.warn("Time {} lays outside the solution range: (t0, t1) = "
                      "({}, {})".format(float(t), m.time0, m.time1))
    # assign time
    m.time = t
    # get solution from dense output
    x = m._sol.sol(t)
    # set solution and update
    linear._set_solution(m, x)
    m.update()
    return m.lhs.copy()


class ODESolver(linear.LinearSolver):
    ode_solver = 'BDF'
    default_options = {'verbose': False}

    def _divide_by_jact(self, v):
        if hasattr(self.model, 'jact'):
            v[:self.nlhs] /= self.model.jact
        if hasattr(self.model, 'jaclmt'):
            v[self.nlhs:] /= self.model.jaclmt
        return v

    def _divide_matrix_by_jact(self, A, jact):
        if scipy.sparse.issparse(A):
            A = A.multiply(1./jact[:, None]).tocsr()
        else:
            A = A/jact[:, None]
        return A

    def _update_K(self):
        super(ODESolver, self)._update_K()
        if self.K is not None and hasattr(self.model, 'jact'):
            self.K = self._divide_matrix_by_jact(self.K, self.model.jact)

    def _update_B(self):
        super(ODESolver, self)._update_B()
        if self.B is not None and hasattr(self.model, 'jaclmt'):
            self.B = self._divide_matrix_by_jact(self.B, self.model.jaclmt)
            self.BT = self.B.T

    @property
    def b(self):
        b = super(ODESolver, self).b
        return self._divide_by_jact(b)

    @property
    def l(self):
        l = super(ODESolver, self).l
        return self._divide_by_jact(l)

    @property
    def r(self):
        # This is faster then simply self.l - self.b
        # because division is made once
        r = super(ODESolver, self).l - super(ODESolver, self).b
        return self._divide_by_jact(r)

    def residual(self, t, x, verbose=True):
        self.update_model(t, x)
        self.x
        r = -self.r
        if self._verbose and verbose:
            print('{:10f}: residual called'.format(t))
        return r

    def jacobian(self, t, x):
        self._update_K()
        self._update_B()
        jac = self.Jsparse
        if jac is None:
            raise ValueError("'%s' solver needs jacobian" % (self.ode_solver))
        if self._verbose:
            print('{:10f}: jacobian updated'.format(t))
        return -jac

    def getopts(self, options):
        options = super(ODESolver, self).getopts(options)
        self._verbose = options['verbose']
        del options['verbose']  # solve_ivp does not accept verbose
        return options

    def solve(self, **options):
        options = self.getopts(options)
        t_span = (self.model.time0, self.model.time1)
        ode_solver = self.ode_solver
        jacobian = self.jacobian
        sol = solve_ivp(self.residual, t_span, self.x,
                        method=ode_solver, jac=jacobian,
                        dense_output=True, **options)
        # Make sure last solution is applied to the model
        self.model._sol = sol
        self.model.__class__.sol = _set_solution_at_time
        self.model.sol()  # This sets sotution at last time step

    def update_model(self, t=None, x=None):
        if t is not None:
            self.model.time = t
        super(ODESolver, self).update_model(x)

    def update(self, t=None, x=None, f=None):
        # assign x to model and call its update if it exists
        self.update_model(t, x)
        # clear current data grab new data from model
        self.clear()
        self.x0 = self.x  # here we set also all shaping data
        self._update_K()
        self._update_B()
        self._update_P()
        self._update_Klm()
        self._update_Mlm()


class BDFSolverDS(ODESolver):
    # ode_solver = 'BDF'
    pass


class RadauSolverDS(ODESolver):
    ode_solver = 'Radau'


# class LSODASolverDS(ODESolver):  # NOT WORKING CURRENTLY...
#    ode_solver = 'LSODA'
#
#    def jacobian(self, t, x):
#        jac = super(LSODASolverDS, self).jacobian(t, x)
#        return jac.toarray()


class BDFSolver(ODESolver):
    ode_solver = _bdf.FempyBDF

    def update(self, t=None, x=None):
        self.update_model(t, x)
        self.clear()
        self.x0 = self.x  # here we set also all shaping data
        self._update_K()
        self._update_B()
        # self._update_P()
        # self._update_Klm()
        # self._update_Mlm()
        self.K0 = self.K
        self.B0 = self.B
        self.BT0 = self.BT
        self.I = _identity_like(self.K0)
        if self.nlm:
            self.Ilm = _identity_like(self.B, n=len(self.B)) 
        if self.K is None:
            self.t0 = self.model.time
            self.y0 = self.x0
            self.f0 = self.residual(self.t0, self.y0)
        if self._verbose:
            print('{:10f}: jacobian updated'.format(self.model.time))

    def update_c(self, c):
        self.c = c
        if self.K is not None:
            self.K = self.I - c * self.K0
            if self.nlm:
                self.B = -c * self.B0
                self.BT = -c * self.BT0
            self._update_P()
            self._update_Klm()
            if self.Klm is not None:
                self.Klm -= self.Ilm  # take Ilm into account
            self._update_Mlm()

    # Unfortunately ne need to hack update_P.
    # This is because our jacobian is now not K, but I - c*K and changes are
    # needed to properly handle bcs and bcseq. Possibly this could be
    # better done (but not here i think, in linear solver rather).
    def _update_P(self):
        K = self.K
        nbcs = self.nbcs
        nbcseq = self.nbcseq
        pre = self.preconditioner if hasattr(self, 'preconditioner') else None
        if K is None or pre is None:
            self.P = None
            return
        if not nbcs and not nbcseq:
            P = pre(K)  # LOP or sp matrix
        else:
            bcsr = self.bcsr.copy()
            c = self.c
            if nbcseq:
                K = K.copy()  # this is safe but this is a copy...
                # TODO: K should not be copied here at all
                c0, c1, c2 = self.bcseqc
                bcseq0 = self.bcseq0
                bcseq1 = self.bcseq1
                c01 = c0/c1 * (-c / (1. - c))  # WE HACK HERE ...
                K[bcseq0] += K[bcseq1]
                K[bcseq0, bcseq1] -= 1.  # ... AND HERE ...
                K[:, bcseq0] -= linear._multiply(K[:, bcseq1], c01)
                bcsr[bcseq1] = True
            K0 = K[~bcsr]
            K0 = K0[:, ~bcsr]
            Kbcs = K[:, bcsr]
            P0 = pre(K0)  # LOP or sp matrix

            def matvec(b):
                b = b.copy()  # need a copy, to not alter original b
                b[bcsr] /= (1. - self.c)  # ... AND HERE
                bbcs = (Kbcs.dot(b[bcsr]))[~bcsr]  # b carries x at bcsr
                b0 = b[~bcsr] - bbcs
                x0 = P0.dot(b0)
                x = b.copy()                     # copies also x at bcsr
                x[~bcsr] = x0
                if nbcseq:
                    x[bcseq1] -= c01 * x[bcseq0]
                return x
            P = linear.LOP(K.shape, dtype=K.dtype, matvec=matvec)  # LOP
        self.P = P

    # Get also consistent J (for I - c*K)
    def _matvecJ(self, x):
        ll = super(self.__class__, self)._matvecJ(x)
        if self.nlm:
            ll[-self.nlm:] += x[-self.nlm:]  # add Ilm*lmbd
        if self.nbcs:
            ll[:self.nlhs][self.bcsr] += -self.c * x[:self.nlhs][self.bcsr]
        if self.nbcseq:
            ll[:self.nlhs][self.bcseq1] *= -self.c
            ll[:self.nlhs][self.bcseq1] += x[self.bcseq1]
            ll[:self.nlhs][self.bcseq0] -= x[self.bcseq1]
        return ll

    # and Jsparse ...
    @property
    def Jsparse(self):
        Js = super(self.__class__, self).Jsparse.tocsc()
        if self.nlm:
            idx = np.arange(self.nlhs, self.nlhs + self.nlm)
            Js[idx, idx] = 1.  # assign Ilm in place of D in Schur
        if self.nbcs:
            Js[self.bcsr, self.bcsr] += -self.c  # bcsr size!
        if self.nbcseq:
            Js[self.bcseq1] *= -self.c
            Js[self.bcseq1, self.bcseq1] += 1.
            Js[self.bcseq0, self.bcseq1] -= 1.
        return Js

    def step_solve(self, db):
        if self.K is None:  # use KrylovJacobian
            def res(dy):
                c = -self.c
                t0 = self.t0
                y0 = self.y0
                f0 = self.f0
                r = self.residual(t0, y0 + dy, verbose=False)
                return dy - c * (r - f0)
            KJ = KrylovJacobian()  # KJ uses lgmres by default
            KJ.setup(np.zeros_like(db), np.zeros_like(db), res)
            dx = KJ.solve(db)
            self.info = 0
            return dx
        J = self.J
        M = self.M
        # opts = self.getopts()  # TODO: proper options handling
        dx, info = self.method.method(J, db, M=M)  # , **opts)
        self.info = info
        return dx

    @property
    def jacobian(self):
        return self


if __name__ == '__main__':
    from fempy.testing import testmodule
    testmodule(__file__, '-v')



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 19:09:58 2016

@author: mwojc
"""
from __future__ import print_function
import numpy as np
import scipy.sparse
from scipy.integrate import BDF
# from scipy.integrate._ivp import bdf as bdf_module
# bdf_module.NEWTON_MAXITER = 10


class FempyBDF(BDF):
    # Here we hack scipy BDF solver to use fempy linear solvers
    # instead of default lu decompositions. We are able now to use
    # iterative solvers within BDF and also we can implement
    # solution method when no jacobian is available
    def __init__(self, fun, t0, y0, t_bound, max_step=np.inf, rtol=1e-3,
                 atol=1e-6, jac=None, jac_sparsity=None, vectorized=False,
                 **extraneous):
        # initialize scipy BDF with dummy jacobian ...
        n = len(y0)
        dummy_jac = scipy.sparse.csc_matrix((n, n))
        super(FempyBDF, self).__init__(fun, t0, y0, t_bound, max_step=max_step,
                                       rtol=rtol, atol=atol, jac=dummy_jac,
                                       jac_sparsity=jac_sparsity,
                                       vectorized=vectorized, **extraneous)

        # ... and then hack lu and lu_solve functions to use fempy jacobian
        self.fempy_jac = jac
        self.njev += 1
        self.fempy_jac.update()  # in case it was not done before

        def jac_wrapped(t, y):
            self.njev += 1
            self.fempy_jac.update(t, y)
            return 1.
        self.jac = jac_wrapped
        self.I = 0.
        self.J = 1.

        def lu(c):  # This will take c as argument if I=0 and J=1
            self.nlu += 1
            self.fempy_jac.update_c(c)
            return self.fempy_jac
        self.lu = lu

        def solve_lu(fempy_jac, db):
            return fempy_jac.step_solve(db)
        self.solve_lu = solve_lu

    def _step_impl(self):
        toreturn = super(FempyBDF, self)._step_impl()
#        nlhs = self.fempy_jac.nlhs
#        bcsr = self.fempy_jac.bcsr
#        ybcsr = self.fempy_jac.x_bcsr
#        self.y[:nlhs][bcsr] = ybcsr
        if self.fempy_jac._verbose:
            print('{:10f}: TIME STEP ACCEPTED'.format(self.t))
        return toreturn



# -*- coding: utf-8 -*-
"""
Created on Sat Oct 13 18:45:49 2018

@author: awitek
"""
import numpy as np
from fempy.models.heatandmoisture import HeatAndMoisture


class Test_HAM(object):
    def test_EN_15026(self):
        from fempy.geometry import Point, Line
        A = Point([0, 0, 0], groups='left', elsize_factor=0.001)
        B = Point([15, 0, 0], groups='right', elsize_factor=0.15)
        AB = Line(A, B, groups='construction')
        d = AB.gmsh.domain(2)

        c = construction()
        c.useT(False)
        mat = [c]
        fsn = ['construction']
        env = [outside()]
        sds = ['left']

        m = HAMvalidation(d, mat, fsn, env, sds)
        # setting adiabatic surfaces
        m.T[:] = 293.15
        m.fi[:] = 0.5
        m.alpha[:] = 1.0e4
        m.beta[:] = 1.0e-3
        m.alpha['~left'] = 0.0
        m.beta['~left'] = 0.0
        m.time0 = 0.              # start time
        m.time1 = 7. * 24 * 3600  # finish time, 7 days
        m.solve(verbose=False, rtol=1e-4, atol=1e-4)  # solve
        ###
        scs = m.domain.ngroups.nums(sortby='x')
        x = m.domain.coors[scs, 0]
        # Test after 7 days
        m.sol(7. * 24 * 3600)
        psuc = -c.R/c.Mw * 293.15 * 1.0e3 * np.log(m.fi)
        n_w = 146./(1. + (8.0e-8 * psuc)**1.6)**0.375
        y = n_w[scs]
        # data from standard EN 15026 - test n_w
        xx = [0.01, 0.02, 0.03, 0.04]
        miny7 = [50.2, 41.3, 40.8, 40.8]
        maxy7 = [54.5, 45.6, 45.1, 45.1]
#        miny30=[[0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.08],
#                [81.0, 51.1, 43.6, 41.5, 40.9, 40.8, 40.8]]
#        maxy30=[[0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.08],
#                [85.3, 55.3, 47.9, 45.7, 45.2, 45.1, 45.1]]
#        miny365=[[0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1],
#                 [117.5, 104.4, 88.7, 75.6, 62.8, 55.7, 47.9, 44.1]]
#        maxy365=[[0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.08, 0.1],
#                 [121.8, 108.7, 93.0, 77.9, 67.1, 60.0, 52.2, 48.4]]
        resfi = np.interp(xx, x, y)
        assert np.all(resfi > miny7)
        assert np.all(resfi < maxy7)
        # data from standard EN 15026  - test temperature
        xx = [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 4.0, 5.0]
        miny7 = [26.4, 23.6, 21.7, 20.6, 20.0, 19.8, 19.8, 19.8]
        maxy7 = [26.9, 24.1, 22.2, 21.1, 20.5, 20.4, 20.3, 20.3]
#        miny30=[x, [28.1, 26.5, 25.0, 23.7, 22.7, 21.8, 20.7, 20.1]]
#        maxy30=[x, [28.6, 27.0, 25.5, 24.3, 23.2, 22.3, 21.2, 20.6]]
#        miny365=[x, [29.2, 28.8, 28.3, 27.8, 27.4, 26.9, 26.0, 25.2]]
#        maxy365=[x, [29.8, 29.3, 28.8, 28.4, 27.9, 27.4, 26.6, 25.7]]
        y = m.T[scs] - 273.15
        resT = np.interp(xx, x, y)
        assert np.all(resT > miny7)
        assert np.all(resT < maxy7)
        # Test specific temperature at some point
        num = 800
        np.testing.assert_almost_equal(m.T[num], 294.370064598995)


class outside():
    t2 = 1*1*3600

    def temp(self, time):
        return 293.15 + 10. * time/max(time, self.t2)

    def fi(self, time):
        return 0.5 + 0.45 * time/max(time, self.t2)


class HAMvalidation(HeatAndMoisture):
    def setup(self, materials, field_names, environment, sides):
        self.materials = materials
        self.field_names = field_names
        self.environment = environment
        self.sides = sides
        # Set constants
        self.rho_w[:] = 1000.
        self.cp_w[:] = 4190.
        # Initialize variables
        self.update()

    def coefficients_update(self):
        from fempy.fields.operators import ongauss
        T = ongauss(self.T)
        fi = ongauss(self.fi)
        for i, m in enumerate(self.materials):
            n = self.field_names[i]
            #
            f = fi[n]
            m.setfi(f)
            temp = T[n]
            m.setT(temp)
            m.useT(True)
            #
            self.rho_s[n] = m.rho
            self.cp_s[n] = m.cp
            self.w[n] = m.w
            self.ksi[n] = m.ksi
            #
            k = m.k
            d = m.vdp
            D = m.dfi
            #
            self.kij[n, ..., 0, 0] = k
            self.dp[n, ..., 0, 0] = d
            self.Dfi[n, ..., 0, 0] = D
            s = self.kij.shape[-1]
            if s > 1:
                self.kij[n, ..., 1, 1] = k
                self.dp[n, ..., 1, 1] = d
                self.Dfi[n, ..., 1, 1] = D
            if s > 2:
                self.kij[n, ..., 2, 2] = k
                self.dp[n, ..., 2, 2] = d
                self.Dfi[n, ..., 2, 2] = D

    def environments_update(self):
        t = self.time
        for i, s in enumerate(self.sides):
            envir = self.environment[i]
            self.Text[s] = envir.temp(t)
            if 'fi' in dir(envir):
                fi = envir.fi(t)
            elif 'press' in dir(envir):
                fi = envir.press(t)/self.pextsat
            self.fiext[s] = fi


class materials():
    Mw = 0.01801528            # molar weight of water [kg/mol]
    R = 8.3144598              # universal gas constant [J/(mol*K)]

    def setfi(self, fi):
        self.fi = fi

    def setT(self, T):
        self.T = T

    def useT(self, yn):
        if yn:
            self.temp = self.T
        else:
            self.temp = 293.15

    @property
    def hpsuc(self):
        return self.psuc / (1.0e3 * 9.81)

    @property
    def psuc(self):
        """
        Moisture storage function
        Standard simplification: constatnt temperature
        """
        f = self.fi
        return - self.R / self.Mw * self.temp * 1.0e3 * np.log(f)

    @property
    def dw(self):
        """ Liquid transport coefficient [m2/s] """
        return - self.KK * self.dpsucdw

    @property
    def dfi(self):
        """ Liquid conduction coefficient [kg/(m*s)] """
        return self.dw * self.ksi


class construction(materials):
    rho = 2280.  # [kg/m3]
    cp = 800.    # [J/(kg*K)]

    @property
    def w(self):
        """ Moisture storage function [kg/m3] """
        return 146. / (1. + (8.0e-8 * self.psuc)**1.6)**0.375

    @property
    def vdp(self):
        """ Vapour permeability of the solid, [kg/(m*s*Pa)],
        accordingly to standard, Wufi splits it to nu and delta"""
        w = self.w
        v = self.Mw / self.R / self.temp
        v *= 26.1e-6 / 200. * (1. - w/146.)
        v /= 0.503 * (1. - w/146.)**2 + 0.497
        return v

    @property
    def k(self):
        """ Thermal conductivity [W/(m*K)]"""
        return 1.5 + 15.8e-3 * self.w  # w - water content

    @property
    def dpsucdw(self):
        """ Suction pressure derivative """
        w = self.w
        return -self.psuc * 0.625 / (1. - (146./w)**(-1./0.375)) / (0.375 * w)

    @property
    def KK(self):
        """ Liquid water conductivity """
        w = self.w
        k = -39.2619 + 0.0704 * (w-73.) - 1.7420e-4 * (w-73.)**2
        k -= 2.7953e-6 * (w-73.)**3 + 1.1566e-7 * (w-73.)**4
        k += 2.5969e-9 * (w-73.)**5
        from numpy import exp
        return exp(k)

    @property
    def ksi(self):
        """ Moisture storage capacity [kg/m3] """
        f = self.fi
        k = 3.867658187008071e-10 * self.R / self.Mw
        k *= self.temp * 1.0e3 * self.psuc**0.6
        k /= f * (4.41513491667588e-12 * self.psuc**1.6 + 1.)**1.375
        return k


if __name__ == '__main__':
    import pytest
    pytest.main([str(__file__), '-v'])



# -*- coding: utf-8 -*-
"""
Created on Fri Jul  4 11:39:16 2014

@author: mwojc
"""
import numpy as np
from fempy.fields.fields import Field


def get_prefixes(a, b):
    if not isinstance(a, Field) and not isinstance(b, Field):
        a = np.asarray(a)
        b = np.asarray(b)
        ap = 0
        af = len(a.shape)
        bp = 0
        bf = len(b.shape)
        rp = 0
    elif isinstance(a, Field) and not isinstance(b, Field):
        b = np.asarray(b)
        ap = a.prefix
        af = a.sufix
        bp = 0
        bf = len(b.shape)
        if len(b.shape) >= ap:
            if b.shape[:ap] == a.pshape:
                bp = ap
                bf = len(b.shape[bp:])
        rp = ap
    elif not isinstance(a, Field) and isinstance(b, Field):
        a = np.asarray(a)
        bp = b.prefix
        bf = b.sufix
        ap = 0
        af = len(a.shape)
        if len(a.shape) >= bp:
            if a.shape[:bp] == b.pshape:
                ap = bp
                af = len(a.shape[ap:])
        rp = bp
    else:
        ap = a.prefix
        af = a.sufix
        bp = b.prefix
        bf = b.sufix
        if a.pshape != b.pshape:
            raise ValueError("Inconsistent fields: %s, %s)"
                             % (a.pshape, b.pshape))
        rp = ap
    return ap, af, bp, bf, rp


def is_matrix_field(a):
    return a.sufix == 2


def is_square_matrix_field(a):
    if is_matrix_field(a):
        return a.sshape[0] == a.sshape[1]
    return False


def is_broadcastable(shp1, shp2):
    for a, b in zip(shp1[::-1], shp2[::-1]):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True
